/////////////////////////////////////////////////////////////////
// InferenceEngine.cc
/////////////////////////////////////////////////////////////////

#include "InferenceEngine.h"

/////////////////////////////////////////////////////////////////
// Multidimensional array indexing
/////////////////////////////////////////////////////////////////

inline int Index2D (int i1, int i2, int n){ return i1*n+i2; }
inline int Index3D (int i1, int i2, int i3, int n){ return (i1*n+i2)*n+i3; }
inline int Index4D (int i1, int i2, int i3, int i4, int n){ return ((i1*n+i2)*n+i3)*n+i4; }

/////////////////////////////////////////////////////////////////
// InferenceEngine::InferenceEngine()
//
// Constructor
/////////////////////////////////////////////////////////////////

InferenceEngine::InferenceEngine (const string &sequence) :
  sequence (sequence), length (sequence.length()-1),
  constraints (sequence.length(), UNKNOWN), 
  inside (NULL), outside (NULL), posterior (NULL),
  viterbi (NULL), traceback (NULL) {
  
  for (int i = 0; i < (int) sequence.length(); i++)
    this->sequence[i] = toupper (sequence[i]);

  allow_range_unpaired = VVI(length+1, VI(length+1, 1));

  // Create character mapping from alphabet to integers
  
  char_mapping = VI(256,alphabet.size());
  for (int i = 0; i < (int) alphabet.size(); i++)
    char_mapping[(unsigned char) alphabet[i]] = i;

  InitializeParameters();
}

/////////////////////////////////////////////////////////////////
// InferenceEngine::~InferenceEngine()
//
// Destructor
/////////////////////////////////////////////////////////////////

InferenceEngine::~InferenceEngine (){
  delete inside; inside = NULL;
  delete outside; outside = NULL;
  delete posterior; posterior = NULL;
  delete viterbi; viterbi = NULL;
  delete traceback; traceback = NULL;
}

/////////////////////////////////////////////////////////////////
// InferenceEngine::InitializeParameters(){
//
// Allocate space for storing parameters.
/////////////////////////////////////////////////////////////////

void InferenceEngine::InitializeParameters(){
#ifdef PARAMS_OUTER_LENGTH
  score_outer_unpaired = LOG_ZERO;
  score_outer_branch = LOG_ZERO;
#endif
#ifdef PARAMS_HELIX_BASE_PAIR
  score_helix_base_pair = VVF(M+1, VF(M+1, LOG_ZERO));
#endif
#ifdef PARAMS_HELIX_CLOSING
  score_helix_closing = VVF(M+1, VF(M+1, LOG_ZERO));
#endif
#ifdef PARAMS_HELIX_LENGTH
  score_helix_length = VF (D+1, LOG_ZERO);
  score_helix_extend = LOG_ZERO;
#endif
#ifdef PARAMS_HELIX_STACKING
  score_helix_stacking = VVVVF (M+1, VVVF(M+1, VVF(M+1, VF(M+1, LOG_ZERO))));
#endif
#ifdef PARAMS_TERMINAL_MISMATCH
  score_terminal_mismatch = VVVVF (M+1, VVVF(M+1, VVF(M+1, VF(M+1, LOG_ZERO))));
#endif
#ifdef PARAMS_HAIRPIN_LENGTH
  score_hairpin_length = VF (C+1, LOG_ZERO);
  score_hairpin_extend = LOG_ZERO;
#endif
#ifdef PARAMS_BULGE_LENGTH
  score_bulge_length = VF (C+1, LOG_ZERO);
#endif
#ifdef PARAMS_INTERNAL_FULL
  score_internal_full = VVF (C+1, VF(C+1, LOG_ZERO));
#endif
#ifdef PARAMS_INTERNAL_LENGTH
  score_internal_length = VF (C+1, LOG_ZERO);
#endif
#ifdef PARAMS_INTERNAL_ASYMMETRY
  score_internal_asymmetry = VF (C+1, LOG_ZERO);
#endif
#ifdef PARAMS_MULTI_LENGTH
  score_multi_base = LOG_ZERO;
  score_multi_unpaired = LOG_ZERO;
  score_multi_paired = LOG_ZERO;
#endif
#ifdef PARAMS_SINGLE_BASE_PAIR_STACKING
  score_single_base_pair_stacking_left = VVVF (M+1, VVF(M+1, VF(M+1, LOG_ZERO)));
  score_single_base_pair_stacking_right = VVVF (M+1, VVF(M+1, VF(M+1, LOG_ZERO)));
#endif
}

/////////////////////////////////////////////////////////////////
// InferenceEngine::InitializeCounts(){
//
// Allocate space for storing expected counts.
/////////////////////////////////////////////////////////////////

void InferenceEngine::InitializeCounts(){
#ifdef PARAMS_OUTER_LENGTH
  counts_outer_unpaired = 0;
  counts_outer_branch = 0;
#endif
#ifdef PARAMS_HELIX_BASE_PAIR
  counts_helix_base_pair = VVF(M+1, VF(M+1, 0));
#endif
#ifdef PARAMS_HELIX_CLOSING
  counts_helix_closing = VVF(M+1, VF(M+1, 0));
#endif
#ifdef PARAMS_HELIX_LENGTH
  counts_helix_length = VF (D+1, 0);
  counts_helix_extend = 0;
#endif
#ifdef PARAMS_HELIX_STACKING
  counts_helix_stacking = VVVVF (M+1, VVVF(M+1, VVF(M+1, VF(M+1, 0))));
#endif
#ifdef PARAMS_TERMINAL_MISMATCH
  counts_terminal_mismatch = VVVVF (M+1, VVVF(M+1, VVF(M+1, VF(M+1, 0))));
#endif
#ifdef PARAMS_HAIRPIN_LENGTH
  counts_hairpin_length = VF (C+1, 0);
  counts_hairpin_extend = 0;
#endif
#ifdef PARAMS_BULGE_LENGTH
  counts_bulge_length = VF (C+1, 0);
#endif
#ifdef PARAMS_INTERNAL_FULL
  counts_internal_full = VVF (C+1, VF(C+1, 0));
#endif
#ifdef PARAMS_INTERNAL_LENGTH
  counts_internal_length = VF (C+1, 0);
#endif
#ifdef PARAMS_INTERNAL_ASYMMETRY
  counts_internal_asymmetry = VF (C+1, 0);
#endif
#ifdef PARAMS_MULTI_LENGTH
  counts_multi_base = 0;
  counts_multi_unpaired = 0;
  counts_multi_paired = 0;
#endif
#ifdef PARAMS_SINGLE_BASE_PAIR_STACKING
  counts_single_base_pair_stacking_left = VVVF (M+1, VVF(M+1, VF(M+1, 0)));
  counts_single_base_pair_stacking_right = VVVF (M+1, VVF(M+1, VF(M+1, 0)));
#endif
}

/////////////////////////////////////////////////////////////////
// InferenceEngine::TransformParameters()
//
// Apply any needed transformations to parameters.
/////////////////////////////////////////////////////////////////

void InferenceEngine::TransformParameters(){

#ifdef PARAMS_HAIRPIN_LENGTH
  for (int i = 1; i <= C; i++)
    score_hairpin_length[i] += score_hairpin_length[i-1];
#endif

#ifdef PARAMS_BULGE_LENGTH
  for (int i = 1; i <= C; i++)
    score_bulge_length[i] += score_bulge_length[i-1];
#endif

#ifdef PARAMS_INTERNAL_LENGTH
  for (int i = 2; i <= C; i++)
    score_internal_length[i] += score_internal_length[i-1];
#endif

#ifdef PARAMS_INTERNAL_ASYMMETRY
  for (int i = 1; i <= C; i++)
    score_internal_asymmetry[i] += score_internal_asymmetry[i-1];
#endif

}

/////////////////////////////////////////////////////////////////
// InferenceEngine::TransformCounts()
//
// Apply any needed transformations to counts.
/////////////////////////////////////////////////////////////////

void InferenceEngine::TransformCounts(){

#ifdef PARAMS_HAIRPIN_LENGTH
  for (int i = 1; i <= C; i++)
    for (int j = 0; j < i; j++)
      counts_hairpin_length[j] += counts_hairpin_length[i];
#endif

#ifdef PARAMS_BULGE_LENGTH
  for (int i = 1; i <= C; i++)
    for (int j = 0; j < i; j++)
      counts_bulge_length[j] += counts_bulge_length[i];
#endif

#ifdef PARAMS_INTERNAL_LENGTH
  for (int i = 2; i <= C; i++)
    for (int j = 1; j < i; j++)
      counts_internal_length[j] += counts_internal_length[i];
#endif

#ifdef PARAMS_INTERNAL_ASYMMETRY
  for (int i = 1; i <= C; i++)
    for (int j = 0; j < i; j++)
      counts_internal_asymmetry[j] += counts_internal_asymmetry[i];
#endif

}

/////////////////////////////////////////////////////////////////
// InferenceEngine::LoadParameters()
//
// Load parameters from parameter object.
/////////////////////////////////////////////////////////////////

void InferenceEngine::LoadParameters(const Parameters &parameters){

#ifdef PARAMS_OUTER_LENGTH
  score_outer_unpaired = parameters.GetParam ("outer_unpaired");
  score_outer_branch = parameters.GetParam ("outer_branch");
#endif

#ifdef PARAMS_HELIX_BASE_PAIR
  for (int i = 0; i < M; i++){
    for (int j = 0; j < M; j++){
      score_helix_base_pair
	[char_mapping[(unsigned char) alphabet[i]]]
	[char_mapping[(unsigned char) alphabet[j]]] =
	parameters.GetParam (SPrintF ("helix_base_pair_%c%c", alphabet[i], alphabet[j]));
    }
  }
#endif

#ifdef PARAMS_HELIX_CLOSING
  for (int i = 0; i < M; i++){
    for (int j = 0; j < M; j++){
      score_helix_closing
	[char_mapping[(unsigned char) alphabet[i]]]
	[char_mapping[(unsigned char) alphabet[j]]] =
	parameters.GetParam (SPrintF ("helix_closing_%c%c", alphabet[i], alphabet[j]));
    }
  }
#endif

#ifdef PARAMS_HELIX_LENGTH
  for (int len = 1; len <= D; len++){
    score_helix_length[len] = parameters.GetParam (SPrintF ("helix_length_%d", len));
  }

  score_helix_extend = parameters.GetParam ("helix_extend");
#endif

#ifdef PARAMS_HELIX_STACKING
  for (int i1 = 0; i1 < M; i1++){
    for (int j1 = 0; j1 < M; j1++){
      for (int i2 = 0; i2 < M; i2++){
	for (int j2 = 0; j2 < M; j2++){
	  score_helix_stacking
	    [char_mapping[(unsigned char) alphabet[i1]]]
	    [char_mapping[(unsigned char) alphabet[j1]]]
	    [char_mapping[(unsigned char) alphabet[i2]]]
	    [char_mapping[(unsigned char) alphabet[j2]]] = 
	    parameters.GetParam (SPrintF ("helix_stacking_%c%c%c%c", 
					  alphabet[i1], alphabet[j1],
					  alphabet[i2], alphabet[j2]));
	}
      }
    }
  }  
#endif

#ifdef PARAMS_TERMINAL_MISMATCH
  for (int i1 = 0; i1 < M; i1++){
    for (int j1 = 0; j1 < M; j1++){
      for (int i2 = 0; i2 < M; i2++){
	for (int j2 = 0; j2 < M; j2++){
	  score_terminal_mismatch
	    [char_mapping[(unsigned char) alphabet[i1]]]
	    [char_mapping[(unsigned char) alphabet[j1]]]
	    [char_mapping[(unsigned char) alphabet[i2]]]
	    [char_mapping[(unsigned char) alphabet[j2]]] =
	    parameters.GetParam (SPrintF ("terminal_mismatch_%c%c%c%c",
					  alphabet[i1], alphabet[j1],
					  alphabet[i2], alphabet[j2]));
	}
      }
    }
  }
#endif

#ifdef PARAMS_HAIRPIN_LENGTH
  for (int i = 0; i <= C; i++){
    score_hairpin_length[i] = parameters.GetParam (SPrintF ("hairpin_length_%d", i));
  }
  score_hairpin_extend = parameters.GetParam ("hairpin_extend");
#endif

#ifdef PARAMS_BULGE_LENGTH
  for (int i = 0; i <= C; i++){
    score_bulge_length[i] = parameters.GetParam (SPrintF ("bulge_length_%d", i));
  }
#endif

#ifdef PARAMS_INTERNAL_FULL
  for (int i = 1; i <= C; i++){
    for (int j = 1; i+j <= C; j++){
      score_internal_full[i][j] = parameters.GetParam (SPrintF ("internal_full_%d_%d", i, j));
    }
  }
#endif

#ifdef PARAMS_INTERNAL_LENGTH
  for (int i = 1; i <= C; i++){
    score_internal_length[i] = parameters.GetParam (SPrintF ("internal_length_%d", i));
  }
#endif

#ifdef PARAMS_INTERNAL_ASYMMETRY
  for (int i = 0; i <= C; i++){
    score_internal_asymmetry[i] = parameters.GetParam (SPrintF ("internal_asymmetry_%d", i));
  }
#endif

#ifdef PARAMS_MULTI_LENGTH
  score_multi_base = parameters.GetParam ("multi_base");
  score_multi_unpaired = parameters.GetParam ("multi_unpaired");
  score_multi_paired = parameters.GetParam ("multi_paired");
#endif

#ifdef PARAMS_SINGLE_BASE_PAIR_STACKING
  for (int i1 = 0; i1 < M; i1++){
    for (int j1 = 0; j1 < M; j1++){
      for (int i2 = 0; i2 < M; i2++){
        score_single_base_pair_stacking_left
          [char_mapping[(unsigned char) alphabet[i1]]]
          [char_mapping[(unsigned char) alphabet[j1]]]
          [char_mapping[(unsigned char) alphabet[i2]]] =
          parameters.GetParam (SPrintF ("single_base_pair_stacking_left_%c%c%c", 
					alphabet[i1], alphabet[j1], alphabet[i2]));
        score_single_base_pair_stacking_right
          [char_mapping[(unsigned char) alphabet[i1]]]
          [char_mapping[(unsigned char) alphabet[j1]]]
          [char_mapping[(unsigned char) alphabet[i2]]] =
          parameters.GetParam (SPrintF ("single_base_pair_stacking_right_%c%c%c", 
					alphabet[i1], alphabet[j1], alphabet[i2]));
      }
    }
  }
#endif

  TransformParameters();
}

/////////////////////////////////////////////////////////////////
// InferenceEngine::WriteCounts()
//
// Write expected counts to parameter object.
/////////////////////////////////////////////////////////////////

void InferenceEngine::WriteCounts(Parameters &counts){

  TransformCounts();

  counts.Clear();

#ifdef PARAMS_OUTER_LENGTH
  counts.AccumulateCounts ("outer_unpaired", counts_outer_unpaired);
  counts.AccumulateCounts ("outer_branch", counts_outer_branch);
#endif

#ifdef PARAMS_HELIX_BASE_PAIR
  for (int i = 0; i < M; i++){
    for (int j = 0; j < M; j++){
      counts.AccumulateCounts (SPrintF ("helix_base_pair_%c%c", alphabet[i], alphabet[j]),
			       counts_helix_base_pair
			       [char_mapping[(unsigned char) alphabet[i]]]
			       [char_mapping[(unsigned char) alphabet[j]]]);
    }
  }
#endif

#ifdef PARAMS_HELIX_CLOSING
  for (int i = 0; i < M; i++){
    for (int j = 0; j < M; j++){
      counts.AccumulateCounts (SPrintF ("helix_closing_%c%c", alphabet[i], alphabet[j]),
			       counts_helix_closing
			       [char_mapping[(unsigned char) alphabet[i]]]
			       [char_mapping[(unsigned char) alphabet[j]]]);
    }
  }
#endif

#ifdef PARAMS_HELIX_LENGTH
  for (int len = 1; len <= D; len++){
    counts.AccumulateCounts (SPrintF ("helix_length_%d", len),
			     counts_helix_length[len]);
  }

  counts.AccumulateCounts ("helix_extend", counts_helix_extend);
#endif

#ifdef PARAMS_HELIX_STACKING
  for (int i1 = 0; i1 < M; i1++){
    for (int j1 = 0; j1 < M; j1++){
      for (int i2 = 0; i2 < M; i2++){
	for (int j2 = 0; j2 < M; j2++){
	  counts.AccumulateCounts (SPrintF ("helix_stacking_%c%c%c%c", 
					    alphabet[i1], alphabet[j1],
					    alphabet[i2], alphabet[j2]),
				   counts_helix_stacking
				   [char_mapping[(unsigned char) alphabet[i1]]]
				   [char_mapping[(unsigned char) alphabet[j1]]]
				   [char_mapping[(unsigned char) alphabet[i2]]]
				   [char_mapping[(unsigned char) alphabet[j2]]]);
	}
      }
    }
  }  
#endif

#ifdef PARAMS_TERMINAL_MISMATCH
  for (int i1 = 0; i1 < M; i1++){
    for (int j1 = 0; j1 < M; j1++){
      for (int i2 = 0; i2 < M; i2++){
	for (int j2 = 0; j2 < M; j2++){
	  counts.AccumulateCounts (SPrintF ("terminal_mismatch_%c%c%c%c",
					    alphabet[i1], alphabet[j1],
					    alphabet[i2], alphabet[j2]),
				   counts_terminal_mismatch
				   [char_mapping[(unsigned char) alphabet[i1]]]
				   [char_mapping[(unsigned char) alphabet[j1]]]
				   [char_mapping[(unsigned char) alphabet[i2]]]
				   [char_mapping[(unsigned char) alphabet[j2]]]);
	}
      }
    }
  }
#endif

#ifdef PARAMS_HAIRPIN_LENGTH
  for (int i = 0; i <= C; i++){
    counts.AccumulateCounts (SPrintF ("hairpin_length_%d", i), counts_hairpin_length[i]);
  }
  counts.AccumulateCounts ("hairpin_extend", counts_hairpin_extend);
#endif

#ifdef PARAMS_BULGE_LENGTH
  for (int i = 0; i <= C; i++){
    counts.AccumulateCounts (SPrintF ("bulge_length_%d", i), counts_bulge_length[i]);
  }
#endif

#ifdef PARAMS_INTERNAL_FULL
  for (int i = 1; i <= C; i++){
    for (int j = 1; i+j <= C; j++){
      counts.AccumulateCounts (SPrintF ("internal_full_%d_%d", i, j), counts_internal_full[i][j]);
    }
  }
#endif

#ifdef PARAMS_INTERNAL_LENGTH
  for (int i = 1; i <= C; i++){
    counts.AccumulateCounts (SPrintF ("internal_length_%d", i), counts_internal_length[i]);
  }
#endif

#ifdef PARAMS_INTERNAL_ASYMMETRY
  for (int i = 0; i <= C; i++){
    counts.AccumulateCounts (SPrintF ("internal_asymmetry_%d", i), counts_internal_asymmetry[i]);
  }
#endif

#ifdef PARAMS_MULTI_LENGTH
  counts.AccumulateCounts ("multi_base", counts_multi_base);
  counts.AccumulateCounts ("multi_unpaired", counts_multi_unpaired);
  counts.AccumulateCounts ("multi_paired", counts_multi_paired);
#endif

#ifdef PARAMS_SINGLE_BASE_PAIR_STACKING
  for (int i1 = 0; i1 < M; i1++){
    for (int j1 = 0; j1 < M; j1++){
      for (int i2 = 0; i2 < M; i2++){
        counts.AccumulateCounts (SPrintF ("single_base_pair_stacking_left_%c%c%c", 
					  alphabet[i1], alphabet[j1], alphabet[i2]),
				 counts_single_base_pair_stacking_left
				 [char_mapping[(unsigned char) alphabet[i1]]]
				 [char_mapping[(unsigned char) alphabet[j1]]]
				 [char_mapping[(unsigned char) alphabet[i2]]]);
        counts.AccumulateCounts (SPrintF ("single_base_pair_stacking_right_%c%c%c", 
					  alphabet[i1], alphabet[j1], alphabet[i2]),
				 counts_single_base_pair_stacking_right
				 [char_mapping[(unsigned char) alphabet[i1]]]
				 [char_mapping[(unsigned char) alphabet[j1]]]
				 [char_mapping[(unsigned char) alphabet[i2]]]);
      }
    }
  }
#endif

}

/////////////////////////////////////////////////////////////////
// InferenceEngine::UseMapping()
//
// Use known secondary structure mapping
/////////////////////////////////////////////////////////////////

void InferenceEngine::UseMapping (const vector<int> &mapping){
  Assert (constraints.size() == mapping.size(), "Contraints and supplied mapping not of the same length!");
  constraints = mapping;
  
  allow_range_unpaired = VVI(length+1, VI(length+1, 1));
  
  // allow_range_unpaired[i][j] = nucleotide i+1 ... nucleotide j can be unpaired

  for (int i = 0; i <= length; i++){
    for (int j = i+1; j <= length; j++){
      allow_range_unpaired[i][j] = 
	allow_range_unpaired[i][j-1] && (constraints[j] == UNKNOWN || constraints[j] == UNPAIRED);
    }    
  }

  delete inside; inside = NULL;
  delete outside; outside = NULL;
  delete posterior; posterior = NULL;
  delete viterbi; viterbi = NULL;
  delete traceback; traceback = NULL;
}

/////////////////////////////////////////////////////////////////
// InferenceEngine::ScoreOuterUnpaired()
// InferenceEngine::CountOuterUnpaired()
//
// Returns the score for nucleotide i belonging to the exterior
// loop.
/////////////////////////////////////////////////////////////////

Score InferenceEngine::ScoreOuterUnpaired (int i) const {
  Assert (1 <= i && i <= length, "Position out-of-range.");
  if (!allow_range_unpaired[i-1][i]) return LOG_ZERO;
  
#ifdef PARAMS_OUTER_LENGTH
  return score_outer_unpaired;
#else
  return 0;
#endif
}

void InferenceEngine::CountOuterUnpaired (int i, Score value){
  Assert (1 <= i && i <= length, "Position out-of-range.");
  if (!allow_range_unpaired[i-1][i]) return;
  
#ifdef PARAMS_OUTER_LENGTH
  counts_outer_unpaired += value;
#endif

}

/////////////////////////////////////////////////////////////////
// InferenceEngine::ScoreOuterBranch()
// InferenceEngine::CountOuterBranch()
//
// Returns the branch score outside of all secondary structure.
/////////////////////////////////////////////////////////////////

Score InferenceEngine::ScoreOuterBranch() const {
#ifdef PARAMS_OUTER_LENGTH
  return score_outer_branch;
#else
  return 0;
#endif
}

void InferenceEngine::CountOuterBranch (Score value){
#ifdef PARAMS_OUTER_LENGTH
  counts_outer_branch += value;
#endif
}

/////////////////////////////////////////////////////////////////
// InferenceEngine::ScoreHelixBasePair()
// InferenceEngine::CountHelixBasePair()
//
// Returns the score for pairing nucleotides i and j.
/////////////////////////////////////////////////////////////////

Score InferenceEngine::ScoreHelixBasePair (int i, int j) const {
  
  Assert (1 <= i && i <= length, "Invalid nucleotide.");
  Assert (1 <= j && j <= length, "Invalid nucleotide.");

  if ((constraints[i] != UNKNOWN && constraints[i] != j) || 
      (constraints[j] != UNKNOWN && constraints[j] != i))
    return LOG_ZERO;

#ifdef PARAMS_HELIX_BASE_PAIR
  return score_helix_base_pair
    [char_mapping[(unsigned char) sequence[i]]]
    [char_mapping[(unsigned char) sequence[j]]];
#else
  return 0;
#endif

}

void InferenceEngine::CountHelixBasePair (int i, int j, Score value){
  
  Assert (1 <= i && i <= length, "Invalid nucleotide.");
  Assert (1 <= j && j <= length, "Invalid nucleotide.");

  if ((constraints[i] != UNKNOWN && constraints[i] != j) || 
      (constraints[j] != UNKNOWN && constraints[j] != i))
    return;

#ifdef PARAMS_HELIX_BASE_PAIR
  counts_helix_base_pair
    [char_mapping[(unsigned char) sequence[i]]]
    [char_mapping[(unsigned char) sequence[j]]] += value;
#endif

}

/////////////////////////////////////////////////////////////////
// InferenceEngine::ScoreHelixClosing()
// InferenceEngine::CountHelixClosing()
//
// Returns the score for pairing nucleotides i and j at the
// terminus of a helix.
/////////////////////////////////////////////////////////////////

Score InferenceEngine::ScoreHelixClosing (int i, int j) const {
  
  Assert (1 <= i && i <= length, "Invalid nucleotide.");
  Assert (1 <= j && j <= length, "Invalid nucleotide.");

  if ((constraints[i] != UNKNOWN && constraints[i] != j) || 
      (constraints[j] != UNKNOWN && constraints[j] != i))
    return LOG_ZERO;

#ifdef PARAMS_HELIX_CLOSING
  return score_helix_closing
    [char_mapping[(unsigned char) sequence[i]]]
    [char_mapping[(unsigned char) sequence[j]]];
#else
  return 0;
#endif

}

void InferenceEngine::CountHelixClosing (int i, int j, Score value){
  
  Assert (1 <= i && i <= length, "Invalid nucleotide.");
  Assert (1 <= j && j <= length, "Invalid nucleotide.");

  if ((constraints[i] != UNKNOWN && constraints[i] != j) || 
      (constraints[j] != UNKNOWN && constraints[j] != i))
    return;

#ifdef PARAMS_HELIX_CLOSING
  counts_helix_closing
    [char_mapping[(unsigned char) sequence[i]]]
    [char_mapping[(unsigned char) sequence[j]]] += value;
#endif

}

/////////////////////////////////////////////////////////////////
// InferenceEngine::ScoreHelixChange()
// InferenceEngine::CountHelixChange()
//
// Returns the score for increasing a helix length from len-1 to
// len.
/////////////////////////////////////////////////////////////////

Score InferenceEngine::ScoreHelixChange (int len) const {
  
  Assert (1 <= len && len <= D, "Invalid helix length change.");

#ifdef PARAMS_HELIX_LENGTH
  return score_helix_length[len];
#else
  return 0;
#endif

}

void InferenceEngine::CountHelixChange (int len, Score value){
  
  Assert (1 <= len && len <= D, "Invalid helix length change.");

#ifdef PARAMS_HELIX_LENGTH
  counts_helix_length[len] += value;
#endif
  
}

/////////////////////////////////////////////////////////////////
// InferenceEngine::ScoreHelixExtend()
// InferenceEngine::CountHelixExtend()
//
// Returns the score for increasing a helix length past D.
/////////////////////////////////////////////////////////////////

Score InferenceEngine::ScoreHelixExtend () const {
  
#ifdef PARAMS_HELIX_LENGTH
  return score_helix_extend;
#else
  return 0;
#endif

}

void InferenceEngine::CountHelixExtend (Score value){

#ifdef PARAMS_HELIX_LENGTH
  counts_helix_extend += value;
#endif
  
}

/////////////////////////////////////////////////////////////////
// InferenceEngine::ScoreHelixStacking()
// InferenceEngine::CountHelixStacking()
//
// Returns the score for stacking nucleotides (i2,j2) on top of
// nucleotides (i1,j1)
/////////////////////////////////////////////////////////////////

Score InferenceEngine::ScoreHelixStacking (int i1, int j1, int i2, int j2) const {
  
  Assert (1 <= i1 && i1 <= length, "Invalid nucleotide.");
  Assert (1 <= j1 && j1 <= length, "Invalid nucleotide.");
  Assert (1 <= i2 && i2 <= length, "Invalid nucleotide.");
  Assert (1 <= j2 && j2 <= length, "Invalid nucleotide.");

  if ((constraints[i1] != UNKNOWN && constraints[i1] != j1) || 
      (constraints[j1] != UNKNOWN && constraints[j1] != i1))
    return LOG_ZERO;

  if ((constraints[i2] != UNKNOWN && constraints[i2] != j2) || 
      (constraints[j2] != UNKNOWN && constraints[j2] != i2))
    return LOG_ZERO;

#ifdef PARAMS_HELIX_STACKING
  return score_helix_stacking
    [char_mapping[(unsigned char) sequence[i1]]]
    [char_mapping[(unsigned char) sequence[j1]]]
    [char_mapping[(unsigned char) sequence[i2]]]
    [char_mapping[(unsigned char) sequence[j2]]];
#else
  return 0;
#endif

}

void InferenceEngine::CountHelixStacking (int i1, int j1, int i2, int j2, Score value){

  Assert (1 <= i1 && i1 <= length, "Invalid nucleotide.");
  Assert (1 <= j1 && j1 <= length, "Invalid nucleotide.");
  Assert (1 <= i2 && i2 <= length, "Invalid nucleotide.");
  Assert (1 <= j2 && j2 <= length, "Invalid nucleotide.");

  if ((constraints[i1] != UNKNOWN && constraints[i1] != j1) || 
      (constraints[j1] != UNKNOWN && constraints[j1] != i1))
    return;

  if ((constraints[i2] != UNKNOWN && constraints[i2] != j2) || 
      (constraints[j2] != UNKNOWN && constraints[j2] != i2))
    return;

#ifdef PARAMS_HELIX_STACKING
  counts_helix_stacking
    [char_mapping[(unsigned char) sequence[i1]]]
    [char_mapping[(unsigned char) sequence[j1]]]
    [char_mapping[(unsigned char) sequence[i2]]]
    [char_mapping[(unsigned char) sequence[j2]]] += value;
#endif

}

/////////////////////////////////////////////////////////////////
// InferenceEngine::ScoreTerminalMismatch()
// InferenceEngine::CountTerminalMismatch()
//
// Returns the score for stacking nucleotides i2 and j2 on top of
// nucleotides (i1,j1) as a terminal mismatch.
/////////////////////////////////////////////////////////////////

Score InferenceEngine::ScoreTerminalMismatch (int i1, int j1, int i2, int j2) const {
  
  Assert (1 <= i1 && i1 <= length, "Invalid nucleotide.");
  Assert (1 <= j1 && j1 <= length, "Invalid nucleotide.");
  Assert (1 <= i2 && i2 <= length, "Invalid nucleotide.");
  Assert (1 <= j2 && j2 <= length, "Invalid nucleotide.");

  if ((constraints[i1] != UNKNOWN && constraints[i1] != j1) || 
      (constraints[j1] != UNKNOWN && constraints[j1] != i1))
    return LOG_ZERO;

#ifdef PARAMS_TERMINAL_MISMATCH
  return score_terminal_mismatch
    [char_mapping[(unsigned char) sequence[i1]]]
    [char_mapping[(unsigned char) sequence[j1]]]
    [char_mapping[(unsigned char) sequence[i2]]]
    [char_mapping[(unsigned char) sequence[j2]]];
#else
  return 0;
#endif

}

void InferenceEngine::CountTerminalMismatch (int i1, int j1, int i2, int j2, Score value){
  
  Assert (1 <= i1 && i1 <= length, "Invalid nucleotide.");
  Assert (1 <= j1 && j1 <= length, "Invalid nucleotide.");
  Assert (1 <= i2 && i2 <= length, "Invalid nucleotide.");
  Assert (1 <= j2 && j2 <= length, "Invalid nucleotide.");

  if ((constraints[i1] != UNKNOWN && constraints[i1] != j1) || 
      (constraints[j1] != UNKNOWN && constraints[j1] != i1))
    return;

#ifdef PARAMS_TERMINAL_MISMATCH
  counts_terminal_mismatch
    [char_mapping[(unsigned char) sequence[i1]]]
    [char_mapping[(unsigned char) sequence[j1]]]
    [char_mapping[(unsigned char) sequence[i2]]]
    [char_mapping[(unsigned char) sequence[j2]]] += value;
#endif

}

/////////////////////////////////////////////////////////////////
// InferenceEngine::ScoreHairpin()
// InferenceEngine::CountHairpin()
//
// Returns the score for a hairpin spanning positions i to j.
/////////////////////////////////////////////////////////////////

Score InferenceEngine::ScoreHairpin (int i, int j) const {
  Assert (0 < i && i <= j && j < length, "Hairpin boundaries invalid.");

  if (!allow_range_unpaired[i][j]) return LOG_ZERO;

  Score score = 0;

#ifdef PARAMS_TERMINAL_MISMATCH
  score += ScoreTerminalMismatch(i, j+1, i+1, j);
#endif

#ifdef PARAMS_HAIRPIN_LENGTH
  score +=
    (j - i <= C) ?
    score_hairpin_length[j - i] :
    score_hairpin_length[C] + (j - i - C) * score_hairpin_extend;
#endif
  
  return score;
}

void InferenceEngine::CountHairpin (int i, int j, Score value){
  Assert (0 < i && i <= j && j < length, "Hairpin boundaries invalid.");

  if (!allow_range_unpaired[i][j]) return;

#ifdef PARAMS_TERMINAL_MISMATCH
  CountTerminalMismatch (i, j+1, i+1, j, value);
#endif
  
#ifdef PARAMS_HAIRPIN_LENGTH
  if (j - i <= C){
    counts_hairpin_length[j - i] += value;
  } else {
    counts_hairpin_length[C] += value;
    counts_hairpin_extend += (j - i - C) * value;
  }
#endif
}

/////////////////////////////////////////////////////////////////
// InferenceEngine::ScoreSingle()
// InferenceEngine::CountSingle()
//
// Returns the score for an internal loop spanning i to j and
// ip to jp.
/////////////////////////////////////////////////////////////////

Score InferenceEngine::ScoreSingle (int i, int j, int ip, int jp) const {
  Assert (ip - i >= 0 && j - jp >= 0, "Invalid single-branch loop size.");
  Assert (1 <= ip - i + j - jp && ip - i + j - jp <= C, "Invalid single-branch loop size.");

  if (!allow_range_unpaired[i][ip]) return LOG_ZERO;
  if (!allow_range_unpaired[jp][j]) return LOG_ZERO;

  Score score = 0;

  if (ip - i == 0 || j - jp == 0){

#ifdef PARAMS_BULGE_LENGTH
    score += score_bulge_length[ip - i + j - jp];
#endif

  } else {
    
#ifdef PARAMS_INTERNAL_FULL
    score += score_internal_full[ip - i][j - jp];
#endif
#ifdef PARAMS_INTERNAL_LENGTH
    score += score_internal_length[ip - i + j - jp];
#endif
#ifdef PARAMS_INTERNAL_ASYMMETRY
    score += score_internal_asymmetry[abs((ip - i) - (j - jp))];
#endif
    
  }

#ifdef PARAMS_TERMINAL_MISMATCH
  score += ScoreTerminalMismatch (i, j+1, i+1, j);
  score += ScoreTerminalMismatch (jp, ip+1, jp+1, ip);
#endif

  return score;
}

void InferenceEngine::CountSingle (int i, int j, int ip, int jp, Score value){
  Assert (ip - i >= 0 && j - jp >= 0, "Invalid single-branch loop size.");
  Assert (1 <= ip - i + j - jp && ip - i + j - jp <= C, "Invalid single-branch loop size.");

  if (!allow_range_unpaired[i][ip]) return;
  if (!allow_range_unpaired[jp][j]) return;

  if (ip - i == 0 || j - jp == 0){

#ifdef PARAMS_BULGE_LENGTH
    counts_bulge_length[ip - i + j - jp] += value;
#endif

  } else {
    
#ifdef PARAMS_INTERNAL_FULL
    counts_internal_full[ip - i][j - jp] += value;
#endif
#ifdef PARAMS_INTERNAL_LENGTH
    counts_internal_length[ip - i + j - jp] += value;
#endif
#ifdef PARAMS_INTERNAL_ASYMMETRY
    counts_internal_asymmetry[abs((ip - i) - (j - jp))] += value;
#endif
    
  }
  
#ifdef PARAMS_TERMINAL_MISMATCH
  CountTerminalMismatch (i, j+1, i+1, j, value);
  CountTerminalMismatch (jp, ip+1, jp+1, ip, value);
#endif
}

/////////////////////////////////////////////////////////////////
// InferenceEngine::ScoreMultiBase()
// InferenceEngine::CountMultiBase()
//
// Returns the base score for a multiloop.
/////////////////////////////////////////////////////////////////

Score InferenceEngine::ScoreMultiBase() const {

#ifdef PARAMS_MULTI_LENGTH
  return score_multi_base;
#else
  return 0;
#endif

}

void InferenceEngine::CountMultiBase (Score value){
#ifdef PARAMS_MULTI_LENGTH
  counts_multi_base += value;
#endif

}

/////////////////////////////////////////////////////////////////
// InferenceEngine::ScoreMultiUnpaired()
// InferenceEngine::CountMultiUnpaired()
//
// Returns the score for an unpaired nucleotide i in a multiloop.
/////////////////////////////////////////////////////////////////

Score InferenceEngine::ScoreMultiUnpaired(int i) const {
  Assert (1 <= i && i <= length, "Position out-of-range.");
  if (!allow_range_unpaired[i-1][i]) return LOG_ZERO;
  
#ifdef PARAMS_MULTI_LENGTH
  return score_multi_unpaired;
#else
  return 0;
#endif

}

void InferenceEngine::CountMultiUnpaired (int i, Score value){
  Assert (1 <= i && i <= length, "Position out-of-range.");
  if (!allow_range_unpaired[i-1][i]) return;
  
#ifdef PARAMS_MULTI_LENGTH
  counts_multi_unpaired += value;
#endif

}

/////////////////////////////////////////////////////////////////
// InferenceEngine::ScoreMultiPaired()
// InferenceEngine::CountMultiPaired()
//
// Returns the score for paired nucleotides in a multiloop.
/////////////////////////////////////////////////////////////////

Score InferenceEngine::ScoreMultiPaired () const {
#ifdef PARAMS_MULTI_LENGTH
  return score_multi_paired;
#else
  return 0;
#endif
}

void InferenceEngine::CountMultiPaired (Score value){
#ifdef PARAMS_MULTI_LENGTH
  counts_multi_paired += value;
#endif
}

/////////////////////////////////////////////////////////////////
// InferenceEngine::ScoreMultiMismatch()
// InferenceEngine::CountMultiMismatch()
//
// Returns the score for a stacking mismatch of i2 and j2 on top
// of i1 and j1 in a multiloop.
/////////////////////////////////////////////////////////////////

Score InferenceEngine::ScoreMultiMismatch (int i1, int j1, int i2, int j2) const {
  Assert (1 <= i1 && i1 <= length, "Invalid nucleotide.");
  Assert (1 <= j1 && j1 <= length, "Invalid nucleotide.");
  Assert (1 <= i2 && i2 <= length, "Invalid nucleotide.");
  Assert (1 <= j2 && j2 <= length, "Invalid nucleotide.");
  
  if ((constraints[i1] != UNKNOWN && constraints[i1] != j1) || 
      (constraints[j1] != UNKNOWN && constraints[j1] != i1))
    return LOG_ZERO;

#ifdef PARAMS_SINGLE_BASE_PAIR_STACKING
  return 
    score_single_base_pair_stacking_left
    [char_mapping[(unsigned char) sequence[i1]]]
    [char_mapping[(unsigned char) sequence[j1]]]
    [char_mapping[(unsigned char) sequence[i2]]] +
    score_single_base_pair_stacking_right
    [char_mapping[(unsigned char) sequence[i1]]]
    [char_mapping[(unsigned char) sequence[j1]]]
    [char_mapping[(unsigned char) sequence[j2]]];
#else
  return 0;
#endif

}

void InferenceEngine::CountMultiMismatch (int i1, int j1, int i2, int j2, Score value){
  Assert (1 <= i1 && i1 <= length, "Invalid nucleotide.");
  Assert (1 <= j1 && j1 <= length, "Invalid nucleotide.");
  Assert (1 <= i2 && i2 <= length, "Invalid nucleotide.");
  Assert (1 <= j2 && j2 <= length, "Invalid nucleotide.");
  
  if ((constraints[i1] != UNKNOWN && constraints[i1] != j1) || 
      (constraints[j1] != UNKNOWN && constraints[j1] != i1))
    return;

#ifdef PARAMS_SINGLE_BASE_PAIR_STACKING
  counts_single_base_pair_stacking_left
    [char_mapping[(unsigned char) sequence[i1]]]
    [char_mapping[(unsigned char) sequence[j1]]]
    [char_mapping[(unsigned char) sequence[i2]]] += value;
  counts_single_base_pair_stacking_right
    [char_mapping[(unsigned char) sequence[i1]]]
    [char_mapping[(unsigned char) sequence[j1]]]
    [char_mapping[(unsigned char) sequence[j2]]] += value;
#endif

}

/////////////////////////////////////////////////////////////////
// InferenceEngine::ComputeInside()
//
// Run inside algorithm.
/////////////////////////////////////////////////////////////////

void InferenceEngine::ComputeInside() {
  delete inside; inside = NULL;
  delete outside; outside = NULL;
  delete posterior; posterior = NULL;

  inside = new Table (NUM_STATES, length+1);
  inside->Fill (LOG_ZERO);

  // Consider subsequences of length m = 0, 1, 2, ..., L 
  
  for (int m = 0; m <= length; m++){
    for (int i = 0, j = m; j <= length; i++, j++){

      // DO_HELIX 0
      
      if (i+2 <= j)
        LOG_PLUS_EQUALS ((*inside)(DO_HELIX,i,j),
			 ScoreHelixChange(1) +
			 ScoreHelixClosing(i+1,j) +
			 ScoreHelixBasePair(i+1,j) +
			 (*inside)(DO_HELIX+1,i+1,j-1));
      
      // OUTER
      
      if (j == length){
	if (i == length)
	  LOG_PLUS_EQUALS ((*inside)(DO_OUTER,i,j), 
			   0);
	if (i < length)
	  LOG_PLUS_EQUALS ((*inside)(DO_OUTER,i,j), 
			   ScoreOuterUnpaired(i+1) + 
			   (*inside)(DO_OUTER,i+1,j));

	for (int ip = i+2; ip <= length; ip++)
	  LOG_PLUS_EQUALS ((*inside)(DO_OUTER,i,j), 
			   ScoreOuterBranch() + 
			   (*inside)(DO_HELIX,i,ip) + 
			   (*inside)(DO_OUTER,ip,j));
      }
      
      // DO_MULTI

      if (i == j) 
	LOG_PLUS_EQUALS ((*inside)(DO_MULTI+2,i,j), 0);

      for (int n = 0; n <= 2; n++){
	if (i < j) 
	  LOG_PLUS_EQUALS ((*inside)(DO_MULTI+n,i,j), 
			   ScoreMultiUnpaired(i+1) +
			   (*inside)(DO_MULTI+n,i+1,j));

	if (i > 0 && j < length){
	  for (int jp = i+2; jp <= j; jp++){
	    LOG_PLUS_EQUALS ((*inside)(DO_MULTI+n,i,j),
			     ScoreMultiPaired() + 
			     ScoreMultiMismatch(jp,i+1,jp+1,i) +
			     (*inside)(DO_HELIX,i,jp) +
			     (*inside)(DO_MULTI+min(2,n+1),jp,j));
	  }
	}
      }

      // DO_LOOP

      if (i > 0 && j < length){
	LOG_PLUS_EQUALS ((*inside)(DO_LOOP,i,j),
			 ScoreHairpin(i,j));
	
	for (int li = 0; li <= C; li++){
	  for (int lj = 0; li+lj <= C; lj++){
	    if (li+lj > 0){
	      int ip = i+li;
	      int jp = j-lj;
	      if (ip+2 <= jp){
		LOG_PLUS_EQUALS ((*inside)(DO_LOOP,i,j),
				 ScoreSingle(i,j,ip,jp) + 
				 (*inside)(DO_HELIX,ip,jp));
	      }
	    }
	  }
	}
	
	if (i+2 <= j)
	  LOG_PLUS_EQUALS ((*inside)(DO_LOOP,i,j),
			   ScoreMultiBase() +
			   ScoreMultiPaired() +
			   ScoreMultiMismatch(i,j+1,i+1,j) +
			   (*inside)(DO_MULTI,i,j));
      }
      
      // DO_HELIX

      if (i > 0 && j < length){
	for (int n = 1; n <= D; n++){
	  if (i+2 <= j){
	    if (n < D)
	      LOG_PLUS_EQUALS ((*inside)(DO_HELIX+n,i,j),
			       ScoreHelixChange(n+1) +
			       ScoreHelixStacking(i,j+1,i+1,j) +
			       ScoreHelixBasePair(i+1,j) +
			       (*inside)(DO_HELIX+n+1,i+1,j-1));
	    else
	      LOG_PLUS_EQUALS ((*inside)(DO_HELIX+n,i,j),
			       ScoreHelixExtend() +
			       ScoreHelixStacking(i,j+1,i+1,j) +
			       ScoreHelixBasePair(i+1,j) +
			       (*inside)(DO_HELIX+D,i+1,j-1));
	  }
	  LOG_PLUS_EQUALS ((*inside)(DO_HELIX+n,i,j),
			   ScoreHelixClosing(j+1,i) +
			   (*inside)(DO_LOOP,i,j));
	}
      }
    }
  }
}

/////////////////////////////////////////////////////////////////
// InferenceEngine::ComputeOutside()
// 
// Run outside algorithm.
/////////////////////////////////////////////////////////////////

void InferenceEngine::ComputeOutside() {
  Assert (inside, "Must compute inside matrix before computing outside matrix.");
  delete outside; outside = NULL;
  delete posterior; posterior = NULL;

  outside = new Table (NUM_STATES, length+1);
  outside->Fill (LOG_ZERO);

  // Consider subsequences of length m = L, L-1, L-2, ..., 0
  
  for (int m = length; m >= 0; m--){
    
    for (int i = 0, j = m; j <= length; i++, j++){

      // DO_HELIX
      
      if (i > 0 && j < length)
	LOG_PLUS_EQUALS ((*outside)(DO_HELIX+1,i,j),
			 ScoreHelixChange(1) +
			 ScoreHelixClosing(i,j+1) +
			 ScoreHelixBasePair(i,j+1) +
			 (*outside)(DO_HELIX,i-1,j+1));

      if (i > 1 && j < length-1){
	for (int n = 2; n <= D; n++){
          LOG_PLUS_EQUALS ((*outside)(DO_HELIX+n,i,j),
			   ScoreHelixChange(n) +
			   ScoreHelixStacking(i-1,j+2,i,j+1) +
			   ScoreHelixBasePair(i,j+1) +
			   (*outside)(DO_HELIX+n-1,i-1,j+1));
	}
	LOG_PLUS_EQUALS ((*outside)(DO_HELIX+D,i,j),
			 ScoreHelixExtend() +
			 ScoreHelixStacking(i-1,j+2,i,j+1) +
			 ScoreHelixBasePair(i,j+1) +
			 (*outside)(DO_HELIX+D,i-1,j+1));
      }

      // DO_LOOP
      
      if (i > 0 && j < length){
	for (int np = 1; np <= D; np++){
	  LOG_PLUS_EQUALS ((*outside)(DO_LOOP,i,j),
			   ScoreHelixClosing(j+1,i) +
			   (*outside)(DO_HELIX+np,i,j));
	}
      }

      // DO_MULTI

      if (i > 0 && j < length){
	if (i+2 <= j){
	  LOG_PLUS_EQUALS ((*outside)(DO_MULTI,i,j),
			   ScoreMultiBase() +
			   ScoreMultiPaired() +
			   ScoreMultiMismatch(i,j+1,i+1,j) +
			   (*outside)(DO_LOOP,i,j));
	}
	for (int n = 0; n <= 2; n++)
	  LOG_PLUS_EQUALS ((*outside)(DO_MULTI+n,i,j),
			   ScoreMultiUnpaired(i) +
			   (*outside)(DO_MULTI+n,i-1,j));
	if (i > 2){
	  for (int n = 1; n <= 2; n++)
	    for (int ip = 1; ip+2 <= i; ip++)
	      LOG_PLUS_EQUALS ((*outside)(DO_MULTI+n,i,j),
			       ScoreMultiPaired() +
			       ScoreMultiMismatch(i,ip+1,i+1,ip) +
			       (*inside)(DO_HELIX,ip,i) +
			       (*outside)(DO_MULTI+n-1,ip,j));
	  for (int ip = 1; ip+2 <= i; ip++)
	    LOG_PLUS_EQUALS ((*outside)(DO_MULTI+2,i,j),
			     ScoreMultiPaired() +
			     ScoreMultiMismatch(i,ip+1,i+1,ip) +
			     (*inside)(DO_HELIX,ip,i) +
			     (*outside)(DO_MULTI+2,ip,j));
	}
      }

      // DO_OUTER

      if (j == length){
	if (i == 0)
	  LOG_PLUS_EQUALS ((*outside)(DO_OUTER,i,j), 
			   0);
	if (i > 0)
	  LOG_PLUS_EQUALS ((*outside)(DO_OUTER,i,j), 
			   ScoreOuterUnpaired(i) + 
			   (*outside)(DO_OUTER,i-1,j));

	if (i >= 2){
	  for (int ip = 0; ip+2 <= i; ip++)
	    LOG_PLUS_EQUALS ((*outside)(DO_OUTER,i,j), 
			     ScoreOuterBranch() + 
			     (*inside)(DO_HELIX,ip,i) + 
			     (*outside)(DO_OUTER,ip,j));
	}
      }
      
      // DO_HELIX 0
      
      if (i+2 <= j){
	LOG_PLUS_EQUALS ((*outside)(DO_HELIX,i,j),
			 ScoreOuterBranch() + 
			 (*outside)(DO_OUTER,i,length) + 
			 (*inside)(DO_OUTER,j,length));

	if (i > 0 && j < length){
	  for (int li = 0; li <= C; li++){
	    for (int lj = 0; li+lj <= C; lj++){
	      if (li+lj > 0){
		int ip = i-li;
		int jp = j+lj;
		if (ip > 0 && jp < length){		  
		  LOG_PLUS_EQUALS ((*outside)(DO_HELIX,i,j),
				   ScoreSingle(ip,jp,i,j) + 
				   (*outside)(DO_LOOP,ip,jp));
		}
	      }
	    }
	  }
	  
	  for (int np = 0; np <= 2; np++){
	    for (int jp = j; jp < length; jp++){
	      LOG_PLUS_EQUALS ((*outside)(DO_HELIX,i,j),
			       ScoreMultiPaired() +
			       ScoreMultiMismatch(j,i+1,j+1,i) +
			       (*inside)(DO_MULTI+min(2,np+1),j,jp) +
			       (*outside)(DO_MULTI+np,i,jp));
	    }
	  }
	}
      }
    }
  }

}

/////////////////////////////////////////////////////////////////
// InferenceEngine::ComputePartitionCoefficient()
// 
// Compute partition coefficient.  This function requires that
// the inside algorithm has already been called.
/////////////////////////////////////////////////////////////////

Score InferenceEngine::ComputePartitionCoefficient() const {  
  Assert (inside, "Must compute inside matrix before computing partition coefficient.");
  
  //cerr << "inside " << (*inside)(DO_OUTER,0,length) 
    //<< " outside " << (*outside)(DO_OUTER,length,length) 
  //<< endl;
  //exit(1);
 
  return (*inside)(DO_OUTER,0,length);
}

/////////////////////////////////////////////////////////////////
// InferenceEngine::ComputePosterior()
// 
// Combine the results of the inside and outside algorithms
// in order to compute posterior pairing/unpairing probabilities 
// for each letter.
/////////////////////////////////////////////////////////////////

void InferenceEngine::ComputePosterior (){
  Assert (inside, "Must compute inside matrix before computing posterior matrix.");
  Assert (outside, "Must compute outside matrix before computing posterior matrix.");

  delete posterior; posterior = NULL;

  posterior = new Table (1, length+1);
  posterior->Fill (LOG_ZERO);

  const Score Z = ComputePartitionCoefficient();

  for (int i = 1; i <= length; i++){
    for (int j = i+1; j <= length; j++){
      LOG_PLUS_EQUALS ((*posterior)(0,i,j),
		       ScoreHelixChange(1) +
		       ScoreHelixClosing(i,j) +
		       ScoreHelixBasePair(i,j) +
		       (*inside)(DO_HELIX+1,i,j-1) +
		       (*outside)(DO_HELIX,i-1,j));
      
      if (i > 1 && j < length){
	for (int n = 2; n <= D; n++)
	  LOG_PLUS_EQUALS ((*posterior)(0,i,j),
			   ScoreHelixChange(n) +
			   ScoreHelixStacking(i-1,j+1,i,j) +
			   ScoreHelixBasePair(i,j) +
			   (*inside)(DO_HELIX+n,i,j-1) +
			   (*outside)(DO_HELIX+n-1,i-1,j));
	
	LOG_PLUS_EQUALS ((*posterior)(0,i,j),
			 ScoreHelixExtend() +
			 ScoreHelixStacking(i-1,j+1,i,j) +
			 ScoreHelixBasePair(i,j) +
			 (*inside)(DO_HELIX+D,i,j-1) +
			 (*outside)(DO_HELIX+D,i-1,j));
      }
    }
  }
  
  for (int i = 0; i <= length; i++){
    for (int j = i; j <= length; j++){
      (*posterior)(0,i,j) = EXP((*posterior)(0,i,j) - Z);
    }
  }
}
  
/////////////////////////////////////////////////////////////////
// InferenceEngine::ComputeFeatureCountExpectations()
// 
// Combine the results of the inside and outside algorithms
// in order to compute feature count expectations.
/////////////////////////////////////////////////////////////////

void InferenceEngine::ComputeFeatureCountExpectations (vector<double> &feature_expectations){
  Assert (inside, "Must compute inside matrix before computing posterior matrix.");
  Assert (outside, "Must compute outside matrix before computing posterior matrix.");

  InitializeCounts();

  const Score Z = ComputePartitionCoefficient();
  
  // Consider subsequences of length m = 0, 1, 2, ..., L 
  
  for (int m = 0; m <= length; m++){
    for (int i = 0, j = m; j <= length; i++, j++){

      // DO_HELIX 0
      
      if (i+2 <= j){
	Score value = 
	  EXP((*outside)(DO_HELIX,i,j) +
	      ScoreHelixChange(1) +
	      ScoreHelixClosing(i+1,j) +
	      ScoreHelixBasePair(i+1,j) +
	      (*inside)(DO_HELIX+1,i+1,j-1) - Z);

	CountHelixChange(1,value);
	CountHelixClosing(i+1,j,value);
	CountHelixBasePair(i+1,j,value);
      }
      
      // OUTER
      
      if (j == length){
	if (i < length){
	  Score value = 
	    EXP((*outside)(DO_OUTER,i,j) +
		ScoreOuterUnpaired(i+1) +
		(*inside)(DO_OUTER,i+1,j) - Z);
	  
	  CountOuterUnpaired(i+1,value);
	}
	for (int ip = i+2; ip <= length; ip++){
	  Score value = 
	    EXP((*outside)(DO_OUTER,i,j) +
		ScoreOuterBranch() + 
		(*inside)(DO_HELIX,i,ip) + 
		(*inside)(DO_OUTER,ip,j) - Z);
	  
	  CountOuterBranch(value);
	}
      }

      // DO_MULTI

      for (int n = 0; n <= 2; n++){
	if (i < j){
	  Score value = 
	    EXP((*outside)(DO_MULTI+n,i,j) +
		ScoreMultiUnpaired(i+1) +
		(*inside)(DO_MULTI+n,i+1,j) - Z);
	  
	  CountMultiUnpaired(i+1,value);
	}
	
	if (i > 0 && j < length){
	  for (int jp = i+2; jp <= j; jp++){
	    Score value = 
	      EXP((*outside)(DO_MULTI+n,i,j) +
		  ScoreMultiPaired() + 
		  ScoreMultiMismatch(jp,i+1,jp+1,i) +
		  (*inside)(DO_HELIX,i,jp) +
		  (*inside)(DO_MULTI+min(2,n+1),jp,j) - Z);
	    
	    CountMultiPaired(value);
	    CountMultiMismatch(jp,i+1,jp+1,i,value);
	  }
	}
      }
      
      // DO_LOOP

      if (i > 0 && j < length){
	Score value = 
	  EXP((*outside)(DO_LOOP,i,j) +
	      ScoreHairpin(i,j) - Z);
	
	CountHairpin(i,j,value);
	
	for (int li = 0; li <= C; li++){
	  for (int lj = 0; li+lj <= C; lj++){
	    if (li+lj > 0){
	      int ip = i+li;
	      int jp = j-lj;
	      if (ip+2 <= jp){
		Score value = 
		  EXP((*outside)(DO_LOOP,i,j) +
		      ScoreSingle(i,j,ip,jp) + 
		      (*inside)(DO_HELIX,ip,jp) - Z);

		CountSingle(i,j,ip,jp,value);
	      }
	    }
	  }
	}
	
	if (i+2 <= j){
	  value = 
	    EXP((*outside)(DO_LOOP,i,j) +
		ScoreMultiBase() +
		ScoreMultiPaired() +
		ScoreMultiMismatch(i,j+1,i+1,j) +
		(*inside)(DO_MULTI,i,j) - Z);
	  
	  CountMultiBase(value);
	  CountMultiPaired(value);
	  CountMultiMismatch(i,j+1,i+1,j,value);
	}
      }

      // DO_HELIX

      if (i > 0 && j < length){
	for (int n = 1; n <= D; n++){

	  if (i+2 <= j){
	    if (n < D){
	      Score value = 
		EXP((*outside)(DO_HELIX+n,i,j) +
		    ScoreHelixChange(n+1) +
		    ScoreHelixStacking(i,j+1,i+1,j) +
		    ScoreHelixBasePair(i+1,j) +
		    (*inside)(DO_HELIX+n+1,i+1,j-1) - Z);
	      
	      CountHelixChange(n+1,value);
	      CountHelixStacking(i,j+1,i+1,j,value);
	      CountHelixBasePair(i+1,j,value);
	    } else {
	      Score value = 
		EXP((*outside)(DO_HELIX+n,i,j) +
		    ScoreHelixExtend() +
		    ScoreHelixStacking(i,j+1,i+1,j) +
		    ScoreHelixBasePair(i+1,j) +
		    (*inside)(DO_HELIX+D,i+1,j-1) - Z);
	      
	      CountHelixExtend(value);
	      CountHelixStacking(i,j+1,i+1,j,value);
	      CountHelixBasePair(i+1,j,value);
	    }
	  }
	  
	  Score value = 
	    EXP((*outside)(DO_HELIX+n,i,j) +
		ScoreHelixClosing(j+1,i) +
		(*inside)(DO_LOOP,i,j) - Z);
	  
	  CountHelixClosing(j+1,i,value);
	}
      }
    }
  }

  Parameters counts;
  WriteCounts (counts);
  feature_expectations = counts.GetValues();  
}

/////////////////////////////////////////////////////////////////
// InferenceEngine::PredictPairingsPosterior()
// 
// Use posterior decoding to predict pairings.
/////////////////////////////////////////////////////////////////

VI InferenceEngine::PredictPairingsPosterior (const double gammar) const {

  Assert (posterior, "Must compute posterior matrix before predicting pairings via posterior decoding.");

  VVF score(length+1, VF(length+1));
  VF unpaired_posterior(length+1,1);
  VF dp_candidates(length + 4);
  VVI traceback(length+1,VI(length+1));
  queue< pair<int, int> > traceback_queue; // for following traceback
  VI solution(length+1,UNPAIRED);
  solution[0] = UNKNOWN;
  
  const int TB_ZERO = 0;
  const int TB_LEFT = 1;
  const int TB_RIGHT = 2;
  const int TB_PAIR = 3;
  const int TB_BRANCH = 4; // note, add branch offset to this one
  
  // Populate unpaired_posterior (which was initialized to 1's) 
  
  for (int i = 1; i <= length; i++) {
    for (int j = 1; j < i; j++) {
      unpaired_posterior[i] -= (*posterior)(0,j,i);
    }
    for (int j = i + 1; j <= length; j++) {
      unpaired_posterior[i] -= (*posterior)(0,i,j);
    }
  }
  
  
  for (int m = 0; m <= length; m++){
    for (int i = 0, j = m; j <= length; i++, j++){
      
      // Find all possible of score at i,j 
      dp_candidates.resize(4); // 4 is minimum size it will be
      
      dp_candidates[TB_ZERO] = (i == j) ? 0 : LOG_ZERO; // LOG_ZERO means -inf
      
      dp_candidates[TB_LEFT] = (i < j) ?
	unpaired_posterior[i+1] + score[i+1][j] : LOG_ZERO;
      
      dp_candidates[TB_RIGHT] = (i < j) ?
	unpaired_posterior[j] + score[i][j-1] : LOG_ZERO;
      
      dp_candidates[TB_PAIR] = (i + 1 < j) ?
	2 * gammar * (*posterior)(0,i+1,j) + score[i+1][j-1] : LOG_ZERO;
      
      for(int k = i + 1; k < j; k++) { // TB_BRANCH
	dp_candidates.push_back(score[i][k] + score[k][j]);
      }
      
      // Initialize score impossibly small 
      
      score[i][j] = LOG_ZERO;
      
      // Pick the biggest score 
      
      for(int t = 0; t < (int) dp_candidates.size(); t++){
	if(score[i][j] < dp_candidates[t]){
	  score[i][j] = dp_candidates[t];
	  traceback[i][j] = t;
	}
      }
    }
  }
  
  traceback_queue.push(make_pair(0, length));
  while(!traceback_queue.empty()){
    pair<int, int> t = traceback_queue.front();
    traceback_queue.pop();
    
    switch(traceback[t.first][t.second]){
    case TB_ZERO:
      break;
    case TB_LEFT:
      traceback_queue.push(make_pair(t.first+1,t.second));
      break;
    case TB_RIGHT:
      traceback_queue.push(make_pair(t.first,t.second-1));
      break;
    case TB_PAIR:
      solution[t.first+1] = t.second;
      solution[t.second] = t.first+1;
      traceback_queue.push(make_pair(t.first+1,t.second-1));
      break;
    default: // branch
      traceback_queue.push( make_pair(t.first, t.first + traceback[t.first][t.second] - TB_BRANCH + 1));
      traceback_queue.push( make_pair(t.first + traceback[t.first][t.second] - TB_BRANCH + 1, t.second));
      break;
    }
  }
  
  return solution;
}

/////////////////////////////////////////////////////////////////
// InferenceEngine::UPDATE_MAX()
//
// Replace value and update traceback if provided value is
// larger.
/////////////////////////////////////////////////////////////////

void InferenceEngine::UPDATE_MAX (Score &curr_val, 
				  Score &curr_traceback,
                                  const Score &new_val, 
				  const Score &new_traceback) const {
  if (new_val > curr_val){
    curr_val = new_val;
    curr_traceback = new_traceback;
  }
}

/////////////////////////////////////////////////////////////////
// InferenceEngine::EncodeTraceback()
// InferenceEngine::DecodeTraceback()
//
// Encode and decode traceback as a Score.
/////////////////////////////////////////////////////////////////

Score InferenceEngine::EncodeTraceback (int i, int j) const {
  return (i+1) * (length+2) + (j+1);
}

pair<int,int> InferenceEngine::DecodeTraceback (Score s) const {
  int val = (int) s;
  return make_pair (val / (length+2)-1, val % (length+2)-1);
}

/////////////////////////////////////////////////////////////////
// InferenceEngine::ComputeViterbi()
//
// Run inside algorithm.
/////////////////////////////////////////////////////////////////

void InferenceEngine::ComputeViterbi() {
  delete viterbi; viterbi = NULL;
  delete traceback; traceback = NULL;

  viterbi = new Table (NUM_STATES, length+1);
  viterbi->Fill (LOG_ZERO);

  traceback = new Table (NUM_STATES, length+1);
  traceback->Fill (LOG_ZERO);

  // Consider subsequences of length m = 0, 1, 2, ..., L 
  
  for (int m = 0; m <= length; m++){
    for (int i = 0, j = m; j <= length; i++, j++){

      // DO_HELIX 0
      
      if (i+2 <= j)
        UPDATE_MAX ((*viterbi)(DO_HELIX,i,j),
		    (*traceback)(DO_HELIX,i,j),
		    ScoreHelixChange(1) +
		    ScoreHelixClosing(i+1,j) +
		    ScoreHelixBasePair(i+1,j) +
		    (*viterbi)(DO_HELIX+1,i+1,j-1),
		    EncodeTraceback(0,0));
      
      // OUTER
      
      if (j == length){
	if (i == length)
	  UPDATE_MAX ((*viterbi)(DO_OUTER,i,j),
		      (*traceback)(DO_OUTER,i,j),
		      0,
		      EncodeTraceback(0,0));
	if (i < length)
	  UPDATE_MAX ((*viterbi)(DO_OUTER,i,j),
		      (*traceback)(DO_OUTER,i,j),
		      ScoreOuterUnpaired(i+1) + 
		      (*viterbi)(DO_OUTER,i+1,j),
		      EncodeTraceback(1,0));
	
	for (int ip = i+2; ip <= length; ip++)
	  UPDATE_MAX ((*viterbi)(DO_OUTER,i,j),
		      (*traceback)(DO_OUTER,i,j),
		      ScoreOuterBranch() + 
		      (*viterbi)(DO_HELIX,i,ip) + 
		      (*viterbi)(DO_OUTER,ip,j),
		      EncodeTraceback(ip,j));
      }
      
      // DO_MULTI
      
      if (i == j) 
	UPDATE_MAX ((*viterbi)(DO_MULTI+2,i,j), 
		    (*traceback)(DO_MULTI+2,i,j),
		    0,
		    EncodeTraceback(0,0));
      
      for (int n = 0; n <= 2; n++){
	if (i < j) 
	  UPDATE_MAX ((*viterbi)(DO_MULTI+n,i,j), 
		      (*traceback)(DO_MULTI+n,i,j),
		      ScoreMultiUnpaired(i+1) +
		      (*viterbi)(DO_MULTI+n,i+1,j),
		      EncodeTraceback(1,0));
	
	if (i > 0 && j < length){
	  for (int jp = i+2; jp <= j; jp++){
	    UPDATE_MAX ((*viterbi)(DO_MULTI+n,i,j),
			(*traceback)(DO_MULTI+n,i,j),
			ScoreMultiPaired() + 
			ScoreMultiMismatch(jp,i+1,jp+1,i) +
			(*viterbi)(DO_HELIX,i,jp) +
			(*viterbi)(DO_MULTI+min(2,n+1),jp,j),
			EncodeTraceback(jp,0));
	  }
	}
      }
      
      // DO_LOOP
      
      if (i > 0 && j < length){
	UPDATE_MAX ((*viterbi)(DO_LOOP,i,j),
		    (*traceback)(DO_LOOP,i,j),
		    ScoreHairpin(i,j),
		    EncodeTraceback(0,0));
	
	for (int li = 0; li <= C; li++){
	  for (int lj = 0; li+lj <= C; lj++){
	    if (li+lj > 0){
	      int ip = i+li;
	      int jp = j-lj;
	      if (ip+2 <= jp){
		UPDATE_MAX ((*viterbi)(DO_LOOP,i,j),
			    (*traceback)(DO_LOOP,i,j),
			    ScoreSingle(i,j,ip,jp) + 
			    (*viterbi)(DO_HELIX,ip,jp),
			    EncodeTraceback(ip,jp));
	      }
	    }
	  }
	}
	
	if (i+2 <= j)
	  UPDATE_MAX ((*viterbi)(DO_LOOP,i,j),
		      (*traceback)(DO_LOOP,i,j),
		      ScoreMultiBase() +
		      ScoreMultiPaired() +
		      ScoreMultiMismatch(i,j+1,i+1,j) +
		      (*viterbi)(DO_MULTI,i,j),
		      EncodeTraceback(1,1));
      }
      
      // DO_HELIX
      
      if (i > 0 && j < length){
	for (int n = 1; n <= D; n++){
	  if (i+2 <= j){
	    if (n < D)
	      UPDATE_MAX ((*viterbi)(DO_HELIX+n,i,j),
			  (*traceback)(DO_HELIX+n,i,j),
			  ScoreHelixChange(n+1) +
			  ScoreHelixStacking(i,j+1,i+1,j) +
			  ScoreHelixBasePair(i+1,j) +
			  (*viterbi)(DO_HELIX+n+1,i+1,j-1),
			  EncodeTraceback(0,0));
	    else
	      UPDATE_MAX ((*viterbi)(DO_HELIX+n,i,j),
			  (*traceback)(DO_HELIX+n,i,j),
			  ScoreHelixExtend() +
			  ScoreHelixStacking(i,j+1,i+1,j) +
			  ScoreHelixBasePair(i+1,j) +
			  (*viterbi)(DO_HELIX+D,i+1,j-1),
			  EncodeTraceback(1,0));
	  }
	  UPDATE_MAX ((*viterbi)(DO_HELIX+n,i,j),
		      (*traceback)(DO_HELIX+n,i,j),
		      ScoreHelixClosing(j+1,i) +
		      (*viterbi)(DO_LOOP,i,j),
		      EncodeTraceback(2,0));
	}
      }
    }
  }
}


pair<int,pair<int,int> > InferenceEngine::make_triple (int i, int j, int k) const {
  return make_pair (i, make_pair (j,k));
}

/////////////////////////////////////////////////////////////////
// InferenceEngine::PredictPairingsViterbi()
// 
// Use Viterbi decoding to predict pairings.
/////////////////////////////////////////////////////////////////

VI InferenceEngine::PredictPairingsViterbi () const {

  Assert (viterbi, "Must compute viterbi matrix before predicting pairings via viterbi decoding.");
  Assert (traceback, "Must compute viterbi matrix before predicting pairings via viterbi decoding.");

  VI solution(length+1,UNPAIRED);
  solution[0] = UNKNOWN;
  queue< pair<int, pair<int, int> > > traceback_queue;
  traceback_queue.push (make_triple (DO_OUTER, 0, length));

  while (!traceback_queue.empty()){
    pair<int,pair<int,int> > t = traceback_queue.front();
    traceback_queue.pop();
    int v = t.first;
    int i = t.second.first;
    int j = t.second.second;

    pair<int,int> tb = DecodeTraceback((*traceback)(v,i,j));
    int p = tb.first;
    int k = tb.second;

    switch (v){

    case DO_OUTER:
      if (p == 1 && k == 0){
	traceback_queue.push (make_triple(DO_OUTER,i+1,j));
      } else if (p > 0){
	traceback_queue.push (make_triple(DO_HELIX,i,p));
	traceback_queue.push (make_triple(DO_OUTER,p,j));
      }
      break;
      
    case DO_HELIX:
      traceback_queue.push (make_triple(DO_HELIX+1,i+1,j-1));
      solution[i+1] = j;
      solution[j] = i+1;
      break;
      
    case DO_HELIX1:
#ifdef PARAMS_HELIX_LENGTH
    case DO_HELIX2:
    case DO_HELIX3:
    case DO_HELIX4:
    case DO_HELIX5:
#endif
      if (p == 0){
	traceback_queue.push (make_triple(v+1,i+1,j-1));
	solution[i+1] = j;
	solution[j] = i+1;
      } else if (p == 1){
	traceback_queue.push (make_triple(v,i+1,j-1));
	solution[i+1] = j;
	solution[j] = i+1;
      } else {
	traceback_queue.push (make_triple(DO_LOOP,i,j));
      }
      break;
    case DO_LOOP:
      if (p == 0 && k == 0){
	// hairpin loop
      } else if (p == 1 && k == 1){
	traceback_queue.push (make_triple(DO_MULTI,i,j));
      } else {
	traceback_queue.push (make_triple(DO_HELIX,p,k));
      }
      break;
    case DO_MULTI:
    case DO_MULTI1:
    case DO_MULTI2:
      {
	int n = v - DO_MULTI;
	if (p == 1 && k == 0){
	  traceback_queue.push (make_triple(v,i+1,j));
	} else if (p >= 2){
	  traceback_queue.push (make_triple(DO_HELIX,i,p));
	  traceback_queue.push (make_triple(DO_MULTI+min(2,n+1),p,j));
	}
      }
      break;
    default:
      Assert (false, "Should not get here.");
    };
    
  }
  
  return solution;
}

/////////////////////////////////////////////////////////////////
// InferenceEngine::GetInside()
// InferenceEngine::GetOutside()
// InferenceEngine::GetPosterior()
// InferenceEngine::GetViterbi()
// InferenceEngine::GetTraceback()
//
// Getters.
/////////////////////////////////////////////////////////////////

const Table *InferenceEngine::GetInside() const {
  Assert (inside, "Inside matrix not yet computed.");
  return inside;
}

const Table *InferenceEngine::GetOutside() const {
  Assert (outside, "Outside matrix not yet computed.");
  return outside;
}

const Table *InferenceEngine::GetPosterior() const {
  Assert (posterior, "Posterior matrix not yet computed.");
  return posterior;
}

const Table *InferenceEngine::GetViterbi() const {
  Assert (viterbi, "Viterbi matrix not yet computed.");
  return viterbi;
}

const Table *InferenceEngine::GetTraceback() const {
  Assert (traceback, "Traceback matrix not yet computed.");
  return traceback;
}
