/////////////////////////////////////////////////////////////////
// Minimizer.h
/////////////////////////////////////////////////////////////////

#ifndef MINIMIZER_H
#define MINIMIZER_H

#include <vector>
#include <sstream>
#include <iostream>
#include <iomanip>
#include <cmath>

using namespace std;

/////////////////////////////////////////////////////////////////
// class Minimizer
/////////////////////////////////////////////////////////////////

class Minimizer {

  bool use_preconditioner;
  int N;
  int M;			   
  double term_ratio;
  double alpha;
  double beta;
  double gamma1;
  double gamma2;
  double min_improvement_ratio;
  int max_line_search_evaluations;

  virtual void Report (const vector<double> &theta, int iteration, double objective, double step_length) = 0;
  virtual void Report (const string &s) = 0;
  virtual void ComputeGradient (vector<double> &g, const vector<double> &x) = 0;
  virtual void ComputeHessianDiagonal (vector<double> &h, const vector<double> &x){ h = x; }
  virtual double ComputeFunction  (const vector<double> &x) = 0;

  double LineSearch (const vector<double> &x, 
		     const vector<double> &d, 
		     const vector<double> &g, 
		     double &f_curr);
 public:
  
  Minimizer(bool use_preconditioner,
	    const int    M                           = 100,      // number of previous gradients to remember
	    const double term_ratio                  = 0.00001,  // required ratio of gradient norm to parameter norm for termination
	    const double alpha                       = 0.001,    // minimum improvement ratio for sufficient decrease
	    const double beta                        = 0.5,      // default step size
	    const double gamma1                      = 0.01,     // maximum step size
	    const double gamma2                      = 0.8,      // minimum step size
	    const double min_improvement_ratio       = 1.01,     // minimum improvement ratio after sufficient decrease
	    const int    max_line_search_evaluations = 16);      // maximum number of line search function evaluations
  
  virtual ~Minimizer(){}
  
  double LBFGS (vector<double> &x0,                      // initial guess of solution
		const double f_required = 1e20,          // upper bound on optimum
		const int max_iterations = 1000);        // maximum number of iterations
    
  void ApproximateGradient (vector<double> &g, 
			    const vector<double> &x, 
			    const double EPSILON = 1e-4);

  void SanityCheck (const vector<double> &x, 
		    const vector<double> &g, 
		    int left, int right, int num);
  
};

double DotProduct (const vector<double> &x, const vector<double> &y);
double Norm (const vector<double> &x);
const vector<double> Sqrt (const vector<double> &x);
const vector<double> operator/(const vector<double> &x, const vector<double> &y);
const vector<double> operator+(const vector<double> &x, const vector<double> &y);
const vector<double> operator-(const vector<double> &x, const vector<double> &y);
const vector<double> &operator+= (vector<double> &x, double c);
const vector<double> &operator/= (vector<double> &x, double c);
const vector<double> operator*(const vector<double> &x, double c);
const vector<double> operator/(const vector<double> &x, double c);

#endif
