/////////////////////////////////////////////////////////////////
// OptimizationWrapper.cc
/////////////////////////////////////////////////////////////////

#include "OptimizationWrapper.h"

extern bool toggle_verbose;
extern bool toggle_check;
extern bool toggle_train;
extern bool toggle_partition;
extern bool toggle_viterbi;
extern bool toggle_use_constraints;
extern string output_parens_filename;
extern string output_bpseq_filename;
extern string output_posteriors_filename;

/////////////////////////////////////////////////////////////////
// ProcessingUnit::operator<()
//
// Comparator used to sort processing units according to their
// difficulty (size).
/////////////////////////////////////////////////////////////////

const bool ProcessingUnit::operator< (const ProcessingUnit &rhs) const {
  return (size > rhs.size);
}

/////////////////////////////////////////////////////////////////
// OptimizationWrapper::OptimizationWrapper()
//
// Constructor.
/////////////////////////////////////////////////////////////////

OptimizationWrapper::OptimizationWrapper (const int id, 
					  const int num_procs,
					  const Parameters &base_params) :
  Minimizer(false), id(id), num_procs(num_procs), indent(0), C(0), 
  base_params (base_params), gammar(0),
  starting_time(num_procs), time_used(0), total_time(0){

  if (id == 0 && toggle_train){
    logfile.open ("optimize.log");
    if (logfile.fail()) Error ("Could not open log file for writing.");
  }
}

/////////////////////////////////////////////////////////////////
// OptimizationWrapper::~OptimizationWrapper()
//
// Destructor.
/////////////////////////////////////////////////////////////////

OptimizationWrapper::~OptimizationWrapper (){
  if (id == 0){
    logfile.close();
    
#ifdef MULTI
    for (int i = 1; i < num_procs; i++){
      CommandType command = TERMINATE;
      MPI_Send (&command, 1, MPI_INT, i, 0, MPI_COMM_WORLD);
    }
#endif    
  }
}

/////////////////////////////////////////////////////////////////
// OptimizationWrapper::ComputeCrossValidationLogLoss()
//
// Compute cross-validation logloss.
/////////////////////////////////////////////////////////////////

vector<double> OptimizationWrapper::ComputeCrossValidationLogLoss (const Parameters &initial_params,
								   const vector<ProcessingUnit> &training,
								   const vector<ProcessingUnit> &holdout,
								   double C){
  
  Assert (id == 0, "Routine should only be called by master process.");
  
  ostringstream oss;
  oss << "Estimating cross validation error for regularization parameter value C" 
      << "=2^(" << log(C) / log(2.0) << ")=" << C;
  PrintMessage (oss.str());
  Indent();
  
  // Train on training set
  
  samples = training;
  vector<double> values = initial_params.GetValues();
  this->C = C;
  time_used = total_time = 0;
  double training_logloss = LBFGS (values);
  
  // Test on holdout set
  
  samples = holdout;
  double testing_logloss = ComputeFunction (values) -
    C * params.LoadValues(values).ComputeRegularizationPenalty();
  
  // Restore old settings
  
  Unindent();
  oss.str("");
  oss << "Computed cross-validation logloss of " << testing_logloss;
  PrintMessage (oss.str());

  vector<double> ret;
  ret.push_back (training_logloss);
  ret.push_back (testing_logloss);
  
  return ret;
}

/////////////////////////////////////////////////////////////////
// OptimizationWrapper::CrossValidate()
//
// Use staged holdout cross-validation in order estimate
// regularization constants.
/////////////////////////////////////////////////////////////////

void OptimizationWrapper::CrossValidate (Parameters &initial_params,
                                         vector<string> filenames, 
                                         double holdout_ratio){

  Assert (id == 0, "Routine should only be called by master process.");
  ostringstream oss;

  // Split data into training and holdout sets

  vector<ProcessingUnit> data = GetProcessingUnits (filenames);
  samples = data;

  sort (data.begin(), data.end());
  //random_shuffle (data.begin() + (int)(0.2 * data.size()), data.end());
  //random_shuffle (data.begin(), data.end());
  vector<ProcessingUnit> holdout (data.begin(), data.begin() + (int) (data.size() * holdout_ratio));
  vector<ProcessingUnit> training (data.begin() + (int) (data.size() * holdout_ratio), data.end());

  if (training.size() == 0 || holdout.size() == 0) 
    Error ("Not enough training samples for cross-validation.");

  oss.str("");
  oss << "Starting regularization parameter search";
  PrintMessage (oss.str());
  Indent();
  
  // Use golden section search
  
  const double RATIO = 0.38197;
  double left = -14, right = 14, mid = left + (right - left) * RATIO;
  vector<double> ret = ComputeCrossValidationLogLoss (initial_params, training, holdout, pow (2.0, mid));
  double f_mid = ret[1];
  double f_total = ret[0] + ret[1];
  
  while (right - left > 1){
    oss.str("");
    oss << "Current bracketing: [" << left << " " << mid << " " << right << "]: " << f_mid;
    PrintMessage (oss.str());
    
    // Compute new trial point
    
    double trial = (right - mid > mid - left) ?
      right - RATIO * (right - left) :
      left + RATIO * (right - left);
    ret = ComputeCrossValidationLogLoss (initial_params, training, holdout, pow (2.0, trial));
    double f_trial = ret[1];
    
    // Now, update bracketing
    
    if (f_trial <= f_mid){
      if (trial >= mid) left = mid; else right = mid;
      mid = trial; f_mid = f_trial;
      f_total = ret[0] + ret[1];
    } else {
      if (trial < mid) left = trial; else right = trial;
    }
  }
  
  Unindent();
  oss.str("");
  oss << "Chose regularization parameter of C = " << pow (2.0, mid);
  PrintMessage (oss.str());
  
  // Now, retrain on all data
  oss.str("");
  oss << "Retraining on entire training set...";
  PrintMessage (oss.str());
  
  Indent();
  samples = data;
  this->C = pow (2.0, mid);
  vector<double> values = initial_params.GetValues();
  time_used = total_time = 0;
  LBFGS (values, f_total);
  initial_params.LoadValues (values);
  Unindent();
}

/////////////////////////////////////////////////////////////////
// OptimizationWrapper::Train()
//
// Run gradient-based optimization with regularization.
/////////////////////////////////////////////////////////////////

void OptimizationWrapper::Train (Parameters &initial_params, 
                                 const vector<string> &filenames,
				 double C){
  
  Assert (id == 0, "Routine should only be called by master process.");
  samples = GetProcessingUnits (filenames);
  vector<double> values = initial_params.GetValues();
  this->C = C;

#ifdef SANITY
  vector<double> gradient;
  ComputeGradient (gradient, values);
  cerr << "Gradient norm: " << Norm(gradient) << endl;

  const vector<ParameterGroup> &param_groups = params.GetParameterGroups();
  for (int i = 0; i < (int) param_groups.size(); i++){
    cerr << "Performing sanity-check for parameter group: " << param_groups[i].name << endl;
    SanityCheck (values, gradient, param_groups[i].begin, param_groups[i].end, 5);
  }
#else
  time_used = total_time = 0;
  LBFGS (values);  
  initial_params.LoadValues (values);
#endif
}

/////////////////////////////////////////////////////////////////
// OptimizationWrapper::GetOutputFilename()
//
// Decide on output filename, if any.
/////////////////////////////////////////////////////////////////

string OptimizationWrapper::GetOutputFilename (bool output_to_directory, 
					       const string &input_filename, 
					       const string &output_filename) const {
  if (output_filename == "") return "";
  if (!output_to_directory) return output_filename;
  
  string::size_type separator_pos = input_filename.find_last_of ('/');
  string base_filename = ((separator_pos == string::npos) ? input_filename : input_filename.substr(separator_pos+1));
  return output_filename + "/" + base_filename;
}

/////////////////////////////////////////////////////////////////
// OptimizationWrapper::GetProcessingUnits()
//
// Build list of units to process
/////////////////////////////////////////////////////////////////

vector<ProcessingUnit> OptimizationWrapper::GetProcessingUnits (const vector<string> &filenames) const {
  Assert (id == 0, "Routine should only be called by master process.");

  // create output directory, if needed

  if (filenames.size() > 1){
    if (output_parens_filename != "" && system (("mkdir -p " + output_parens_filename).c_str())) 
      Error ("Unable to create directory for writing predictions.");
    if (output_bpseq_filename != "" && system (("mkdir -p " + output_bpseq_filename).c_str())) 
      Error ("Unable to create directory for writing predictions.");
    if (output_posteriors_filename != "" && system (("mkdir -p " + output_posteriors_filename).c_str())) 
      Error ("Unable to create directory for writing predictions.");
  }
  
  // build processing units

  vector<ProcessingUnit> ret;
  for (int i = 0; i < (int) filenames.size(); i++){
    SStruct sstruct (filenames[i]);
    ret.push_back (ProcessingUnit());
    ret.back().input_filename = filenames[i];
    ret.back().size = (int) pow(sstruct.GetLength(),3.0);
    ret.back().weight = 1.0;
    ret.back().output_parens_filename = GetOutputFilename (filenames.size() > 1, filenames[i], output_parens_filename);
    ret.back().output_bpseq_filename = GetOutputFilename (filenames.size() > 1, filenames[i], output_bpseq_filename);
    ret.back().output_posteriors_filename = GetOutputFilename (filenames.size() > 1, filenames[i], output_posteriors_filename);
  }
  return ret;  
}

#ifdef MULTI

/////////////////////////////////////////////////////////////////
// OptimizationWrapper::RunComputeNode()
//
// Process queries as a compute node.
/////////////////////////////////////////////////////////////////

void OptimizationWrapper::RunComputeNode (){
  Assert (id != 0, "Routine should not be called by master process.");

  while (true){
    MPI_Status status;
    CommandType command;
    MPI_Recv (&command, 1, MPI_INT, 0, 0, MPI_COMM_WORLD, &status);

    if (command == TERMINATE) break;
    if (command == COMPUTE_FUNCTION || 
        command == COMPUTE_GRADIENT ||
        command == PREDICT){
      
      samples.clear();
      samples.push_back (ProcessingUnit());
      
      // Retrieve parameters
      
      vector<double> new_params (params.GetNumValues());
      MPI_Recv (&new_params.front(), new_params.size(), MPI_DOUBLE, 0, 0, MPI_COMM_WORLD, &status);
      
      // Retrieve gamma
      
      MPI_Recv (&gammar, 1, MPI_DOUBLE, 0, 0, MPI_COMM_WORLD, &status);
      
      // Retrieve filenames
      
      int length;
      char *buffer;
      
      MPI_Recv (&length, 1, MPI_INT, 0, 0, MPI_COMM_WORLD, &status);
      buffer = new char[length+1];
      Assert (buffer, "Failed to allocate memory.");
      MPI_Recv (buffer, length, MPI_CHAR, 0, 0, MPI_COMM_WORLD, &status);
      buffer[length] = '\0';
      samples.back().input_filename = buffer;
      delete [] buffer;
      
      MPI_Recv (&length, 1, MPI_INT, 0, 0, MPI_COMM_WORLD, &status);
      buffer = new char[length+1];
      Assert (buffer, "Failed to allocate memory.");
      MPI_Recv (buffer, length, MPI_CHAR, 0, 0, MPI_COMM_WORLD, &status);
      buffer[length] = '\0';
      samples.back().output_parens_filename = buffer;
      delete [] buffer;

      MPI_Recv (&length, 1, MPI_INT, 0, 0, MPI_COMM_WORLD, &status);
      buffer = new char[length+1];
      Assert (buffer, "Failed to allocate memory.");
      MPI_Recv (buffer, length, MPI_CHAR, 0, 0, MPI_COMM_WORLD, &status);
      buffer[length] = '\0';
      samples.back().output_bpseq_filename = buffer;
      delete [] buffer;
      
      MPI_Recv (&length, 1, MPI_INT, 0, 0, MPI_COMM_WORLD, &status);
      buffer = new char[length+1];
      Assert (buffer, "Failed to allocate memory.");
      MPI_Recv (buffer, length, MPI_CHAR, 0, 0, MPI_COMM_WORLD, &status);
      buffer[length] = '\0';
      samples.back().output_posteriors_filename = buffer;
      delete [] buffer;
      
      // Retrieve extra info
      
      MPI_Recv (&samples.back().size, 1, MPI_DOUBLE, 0, 0, MPI_COMM_WORLD, &status);
      MPI_Recv (&samples.back().weight, 1, MPI_DOUBLE, 0, 0, MPI_COMM_WORLD, &status);
      
      // Return output
      
      switch (command){
      case COMPUTE_FUNCTION: 
	{
	  double function_val = ComputeFunction (new_params);
	  MPI_Send (&function_val, 1, MPI_DOUBLE, 0, 0, MPI_COMM_WORLD);
	  break;        
	}
      case COMPUTE_GRADIENT:
	{
	  vector<double> gradient;
	  ComputeGradient (gradient, new_params); 
	  MPI_Send(&gradient.front(), gradient.size(), MPI_DOUBLE, 0, 0, MPI_COMM_WORLD); 
	  break; 
	}
      case PREDICT:
	{
	  double ret_val = 1;
	  Predict (new_params);
	  MPI_Send(&ret_val, 1, MPI_DOUBLE, 0, 0, MPI_COMM_WORLD);
	  break;
	}
      default:
	Assert (false, "Should not get here.");
      }
      
    } else {
      Error ("Unrecognized command during distributed training.");
    }
  }
}

/////////////////////////////////////////////////////////////////
// OptimizationWrapper::RunMasterNode()
//
// Send queries to compute nodes.
/////////////////////////////////////////////////////////////////

const vector<double> OptimizationWrapper::RunMasterNode (CommandType command, vector<double> values){
  Assert (id == 0, "Routine should only be called by master process.");
  
  MPI_Status status;
  vector<int> assignment (num_procs, -1);
  int num_procs_in_use = 1;
  int curr_unit = 0;
  const time_t total_starting_time = time(0);

  sort (samples.begin(), samples.end());    
  assignment[0] = samples.size();
  
  // Initialization
  
  vector<double> ret;
  switch (command){
  case COMPUTE_FUNCTION:
  case PREDICT:
    ret.resize (1);
    break;
  case COMPUTE_GRADIENT:
    ret.resize (values.size());
    break;
  default:
    Assert (false,
	    "RunMasterNode() can only be called with COMPUTE_FUNCTION, "
	    "COMPUTE_GRADIENT, or PREDICT.");
  }
  
  // As long as there are still outstanding processes or work units left to be completed... 
  
  while (num_procs_in_use > 1 || curr_unit < (int) samples.size()){
    
    // While there are free processors and processing units to do, allocate work.
    
    while (num_procs_in_use < num_procs && curr_unit < (int) samples.size()){
      
      // Find first free processor
      
      int proc = 0;
      while (proc < (int) assignment.size() && assignment[proc] >= 0) proc++;
      Assert (proc < (int) assignment.size(), "Expected to find free processor.");

      starting_time[proc] = time(0);

      // Send command 
      
      MPI_Send (&command, 1, MPI_INT, proc, 0, MPI_COMM_WORLD);
      
      // Send parameters
      
      MPI_Send (&values.front(), values.size(), MPI_DOUBLE, proc, 0, MPI_COMM_WORLD);
      
      // Send gamma
      
      MPI_Send (&gammar, 1, MPI_DOUBLE, proc, 0, MPI_COMM_WORLD);
      
      // Send input filename
      
      int length = (int) samples[curr_unit].input_filename.length();
      MPI_Send (&length, 1, MPI_INT, proc, 0, MPI_COMM_WORLD);
      MPI_Send ((void *) samples[curr_unit].input_filename.c_str(), length, MPI_CHAR, proc, 0, MPI_COMM_WORLD);
      
      // Send output filenames
      
      length = (int) samples[curr_unit].output_parens_filename.length();
      MPI_Send (&length, 1, MPI_INT, proc, 0, MPI_COMM_WORLD);
      MPI_Send ((void *) samples[curr_unit].output_parens_filename.c_str(), length, MPI_CHAR, proc, 0, MPI_COMM_WORLD);

      length = (int) samples[curr_unit].output_bpseq_filename.length();
      MPI_Send (&length, 1, MPI_INT, proc, 0, MPI_COMM_WORLD);
      MPI_Send ((void *) samples[curr_unit].output_bpseq_filename.c_str(), length, MPI_CHAR, proc, 0, MPI_COMM_WORLD);

      length = (int) samples[curr_unit].output_posteriors_filename.length();
      MPI_Send (&length, 1, MPI_INT, proc, 0, MPI_COMM_WORLD);
      MPI_Send ((void *) samples[curr_unit].output_posteriors_filename.c_str(), length, MPI_CHAR, proc, 0, MPI_COMM_WORLD);
      
      // Send extra info
      
      MPI_Send (&samples[curr_unit].size, 1, MPI_DOUBLE, proc, 0, MPI_COMM_WORLD);
      MPI_Send (&samples[curr_unit].weight, 1, MPI_DOUBLE, proc, 0, MPI_COMM_WORLD);
      
      num_procs_in_use++;
      assignment[proc] = curr_unit;
      curr_unit++;
    }

    // Accumulate results
    
    vector<double> partial_ret (ret.size());
    MPI_Recv (&partial_ret.front(), partial_ret.size(), MPI_DOUBLE, MPI_ANY_SOURCE, 0, MPI_COMM_WORLD, &status);
    for (int i = 0; i < (int) ret.size(); i++) ret[i] += partial_ret[i];

    num_procs_in_use--;
    assignment[status.MPI_SOURCE] = -1;

    time_used += time(0) - starting_time[status.MPI_SOURCE];
  }

  total_time += (num_procs - 1) * (time(0) - total_starting_time);
  
  return ret;
}

#endif

/////////////////////////////////////////////////////////////////
// OptimizationWrapper::ComputeFunction()
//
// Compute negative log-likelihood
/////////////////////////////////////////////////////////////////

double OptimizationWrapper::ComputeFunction (const vector<double> &values){

  vector<double> ret(1);

  // perform the actual computation if we're either a compute node 
  // or running in single processor mode

  if (id > 0 || num_procs == 1){
    params.LoadValues (values + base_params.GetValues());

    for (int i = 0; i < (int) samples.size(); i++){

      SStruct sstruct (samples[i].input_filename);
      
      // unconditional likelihood
      
      InferenceEngine engine (sstruct.GetSequence());
      engine.LoadParameters (params);
      engine.ComputeInside();
      Score part_coeff = engine.ComputePartitionCoefficient();

      // conditional likelihood 
      
      engine.UseMapping (sstruct.GetMapping());
      engine.ComputeInside();
      Score cond_part_coeff = engine.ComputePartitionCoefficient();

      // accumulate likelihood

      double likelihood = (double)(cond_part_coeff - part_coeff) * samples[i].weight;

      if (likelihood < LOG_ZERO_REDUCED){
	cerr << "No valid parse for file: " << samples[i].input_filename << endl;
      } 

      ret[0] -= likelihood;
    }
  }

  if (id == 0){

#ifdef MULTI
    ret = RunMasterNode (COMPUTE_FUNCTION, values);
#endif
    
    ret[0] += C * params.LoadValues(values).ComputeRegularizationPenalty();

    if (toggle_check) exit(0);
  }

  return ret[0];
}

/////////////////////////////////////////////////////////////////
// OptimizationWrapper::ComputeGradient()
//
// Compute gradient
/////////////////////////////////////////////////////////////////

void OptimizationWrapper::ComputeGradient (vector<double> &gradient, const vector<double> &values){

  gradient.clear();
  gradient.resize(values.size());

  // perform the actual computation if we're either a compute node 
  // or running in single processor mode

  if (id > 0 || num_procs == 1){
    params.LoadValues (values + base_params.GetValues());

    for (int i = 0; i < (int) samples.size(); i++){
      SStruct sstruct (samples[i].input_filename);
      vector<double> feature_expectations;
      vector<double> cond_feature_expectations;

      // unconditional expected counts 
      
      InferenceEngine engine (sstruct.GetSequence());
      engine.LoadParameters (params);
      engine.ComputeInside();
      engine.ComputeOutside();
      engine.ComputeFeatureCountExpectations(feature_expectations);

      // conditional expected counts 
      
      engine.UseMapping (sstruct.GetMapping());
      engine.ComputeInside();
      engine.ComputeOutside();
      engine.ComputeFeatureCountExpectations(cond_feature_expectations);

      // accumulate gradient 
      
      for (int j = 0; j < (int) feature_expectations.size(); j++){
	gradient[j] -= (double) (cond_feature_expectations[j] - feature_expectations[j]) * samples[i].weight;
      }
    }
  }

  if (id == 0){
#ifdef MULTI
    gradient = RunMasterNode (COMPUTE_GRADIENT, values);
#endif

    vector<double> regularization_gradient =
      params.LoadValues(values).ComputeRegularizationGradient();

    for (int i = 0; i < (int) values.size(); i++)
      gradient[i] += C * regularization_gradient[i];
  }
}

/////////////////////////////////////////////////////////////////
// OptimizationWrapper::Report()
//
// Used by L-BFGS optimization to report progress.
/////////////////////////////////////////////////////////////////

void OptimizationWrapper::Report (const vector<double> &values, int iteration, 
				  double objective, double step_length){
  ostringstream oss;
  oss << "optimize.params.iter" << iteration;
  params.LoadValues (values + base_params.GetValues()).WriteToFile (oss.str());
  
  const double logloss = objective - C * params.LoadValues(values).ComputeRegularizationPenalty();
  const double efficiency = (total_time == 0 ? 100.0 : 100.0 * time_used / total_time);
  
  oss.str("");
  oss << "Iteration " << iteration << ": f = " << objective << " (unregularized = "
      << logloss << "), step length = " << step_length << ", efficiency = " << efficiency << "%";
  PrintMessage (oss.str());
}

void OptimizationWrapper::Report (const string &s){
  PrintMessage (s);
}

/////////////////////////////////////////////////////////////////
// OptimizationWrapper::Indent()
// OptimizationWrapper::Unindent()
// OptimizationWrapper::PrintMessage()
//
// Print indented message.
/////////////////////////////////////////////////////////////////

void OptimizationWrapper::Indent(){ indent++; }
void OptimizationWrapper::Unindent(){ indent--; Assert (indent >= 0, "Cannot unindent"); }
void OptimizationWrapper::PrintMessage (const string &s){
  if (id == 0){
    for (int i = 0; i < indent; i++) cerr << "    ";
    for (int i = 0; i < indent; i++) logfile << "    ";
    cerr << s << endl;
    logfile << s << endl;
  }
}

/////////////////////////////////////////////////////////////////
// OptimizationWrapper::Predict()
//
// Perform prediction.
/////////////////////////////////////////////////////////////////

void OptimizationWrapper::Predict (vector<double> values){

  // perform the actual computation if we're either a compute node 
  // or running in single processor mode

  if (id > 0 || num_procs == 1){
    params.LoadValues (values + base_params.GetValues());
    
    for (int i = 0; i < (int) samples.size(); i++){

      // load sequence, with constraints if necessary
      
      SStruct sstruct (samples[i].input_filename);
      InferenceEngine engine (sstruct.GetSequence());
      if (toggle_use_constraints) engine.UseMapping (sstruct.GetMapping());
      engine.LoadParameters (params);

      if (toggle_partition){
	engine.ComputeInside();
	Score part_coeff = engine.ComputePartitionCoefficient();
	cout << "Partition coefficient for \"" << samples[i].input_filename << "\": " << part_coeff << endl;
	continue;
      }

      SStruct *solution = NULL;
      const Table *posteriors = NULL;

      // perform computation

      if (toggle_viterbi){
	engine.ComputeViterbi();
	solution = new SStruct(sstruct.GetSequence(), 
			       engine.PredictPairingsViterbi(), 
			       sstruct.GetName());
      } else {
	engine.ComputeInside();
	engine.ComputeOutside();
	engine.ComputePosterior();
	posteriors = engine.GetPosterior();
	solution = new SStruct(sstruct.GetSequence(), 
			       engine.PredictPairingsPosterior(gammar), 
			       sstruct.GetName());
      }

      // write output

      if (samples[i].output_parens_filename != ""){
	ofstream outfile (samples[i].output_parens_filename.c_str());
	if (outfile.fail()) Error ("Unable to open output parens file for writing.");
	solution->WriteParens (outfile);
	outfile.close();
      }

      if (samples[i].output_bpseq_filename != ""){
	ofstream outfile (samples[i].output_bpseq_filename.c_str());
	if (outfile.fail()) Error ("Unable to open output bpseq file for writing.");
	solution->WriteBPSEQ (outfile);
	outfile.close();
      }
	
      if (samples[i].output_posteriors_filename != ""){
	ofstream outfile (samples[i].output_posteriors_filename.c_str());
	if (outfile.fail()) Error ("Unable to open output posteriors file for writing.");
	posteriors->PrintSparse (outfile, sstruct.GetSequence());
	outfile.close();
      }
      
      if (samples[i].output_parens_filename == "" && 
	  samples[i].output_bpseq_filename == "" && 
	  samples[i].output_posteriors_filename == ""){
	solution->WriteParens (cout);
      }
      
      delete solution;
    }
  }

  if (id == 0){
#ifdef MULTI
    RunMasterNode (PREDICT, values);
#endif
  }
}

/////////////////////////////////////////////////////////////////
// OptimizationWrapper::DoPredictions()
//
// Run prediction algorithm
/////////////////////////////////////////////////////////////////

void OptimizationWrapper::DoPredictions (Parameters &initial_params, 
                                         const vector<string> &filenames,
					 const double gammar){

  Assert (id == 0, "Routine should only be called by master process.");
  samples = GetProcessingUnits (filenames);
  this->gammar = gammar;
  
  Predict (initial_params.GetValues());
}
