/////////////////////////////////////////////////////////////////
// OptimizationWrapper.h
/////////////////////////////////////////////////////////////////

#ifdef MULTI
#include <mpi.h>
#endif

#include <fstream>
#include <sstream>
#include <string>
#include <vector>
#include <algorithm>
#include "Utilities.h"
#include "Minimizer.h"
#include "Parameters.h"
#include "SStruct.h"
#include "InferenceEngine.h"

using namespace std;

/////////////////////////////////////////////////////////////////
// struct ProcessingUnit
/////////////////////////////////////////////////////////////////

struct ProcessingUnit {
  string input_filename;
  string output_parens_filename;
  string output_bpseq_filename;
  string output_posteriors_filename;
  int size;
  double weight;
  const bool operator< (const ProcessingUnit &rhs) const;
};

/////////////////////////////////////////////////////////////////
// class OptimizationWrapper
//
// Wrapper class for Minimizer.
/////////////////////////////////////////////////////////////////

class OptimizationWrapper : public Minimizer {
  
  enum CommandType { 
    COMPUTE_FUNCTION,
    COMPUTE_GRADIENT, 
    PREDICT,
    TERMINATE
  };

  int id;
  int num_procs;
  int indent;
  double C;
  ofstream logfile;
  const Parameters &base_params;
  Parameters params;
  vector<ProcessingUnit> samples;
  vector<double> ret;
  double gammar;

  vector<time_t> starting_time;
  double time_used;
  double total_time;

  vector<double> ComputeCrossValidationLogLoss (const Parameters &initial_params,
						const vector<ProcessingUnit> &training,
						const vector<ProcessingUnit> &holdout,
						double C);

  string GetOutputFilename (bool output_to_directory, 
			    const string &input_filename, 
			    const string &output_filename) const;

  vector<ProcessingUnit> GetProcessingUnits (const vector<string> &filenames) const;

  double ComputeFunction (const vector<double> &values);
  void ComputeGradient (vector<double> &gradient, const vector<double> &values);
  void Predict (vector<double> values);

  void Report (const vector<double> &values, int iteration, double objective, double step_length);
  void Report (const string &s);
  void PrintMessage (const string &s);
  void Indent();
  void Unindent();

public:

  OptimizationWrapper (const int id, 
		       const int num_procs,
		       const Parameters &base_params);
  ~OptimizationWrapper ();

  void CrossValidate (Parameters &initial_params,
                      vector<string> filenames, 
                      double holdout_ratio);

  void Train (Parameters &initial_params, 
              const vector<string> &filenames,
	      double C);

  void DoPredictions (Parameters &initial_params, 
                      const vector<string> &filenames,
		      const double gammar);

#ifdef MULTI
  void RunComputeNode();
  const vector<double> RunMasterNode (CommandType command, vector<double> values);
#endif

};  
