/////////////////////////////////////////////////////////////////
// PlotRNA.cc
/////////////////////////////////////////////////////////////////

/////////////////////////////////////////////////////////////////
// Header files
/////////////////////////////////////////////////////////////////

#include "gd.h"
#include "gdfontmb.h"
#include "gdfonts.h"
#include <iostream>
#include <vector>
#include <map>
#include <string>
#include <sstream>
#include <fstream>
#include <cstring>
#include <cstdio>
#include <cmath>
#include "Utilities.h"
#include "SStruct.h"

using namespace std;

/////////////////////////////////////////////////////////////////
// Constants
/////////////////////////////////////////////////////////////////

const int HIGH_CONFIDENCE_COLOR[] = { 0, 0, 0 };
const int LOW_CONFIDENCE_COLOR[] = { 216, 216, 255 };

const double SCALE = 12;
const int BORDER = 50;
const int FONTSIZE = 14;
const double PI = 3.141592653589793238462643383279502884197169399;

/////////////////////////////////////////////////////////////////
// Types
/////////////////////////////////////////////////////////////////

struct Point {
  double x, y;
  Point () : x(0), y(0) {}
  Point (double x, double y) : x(x), y(y) {}
};

/////////////////////////////////////////////////////////////////
// Globals
/////////////////////////////////////////////////////////////////

int HEADER = 0;
bool toggle_complementary_only = false;

/////////////////////////////////////////////////////////////////
// ReadCoords()
//
// Read coordinates file.
/////////////////////////////////////////////////////////////////

vector<Point> ReadCoords (const string &filename){
  vector<Point> res (1);
  ifstream infile (filename.c_str());
  if (infile.fail()) Error ("Could not read coordinates file.");

  double first, second;
  while (infile >> first >> second)
    res.push_back (Point(first, second));
  
  infile.close();
  return res;  
}

/////////////////////////////////////////////////////////////////
// ReadPosteriors()
//
// Read posteriors.
/////////////////////////////////////////////////////////////////

vector<double> ReadPosteriors (const string &filename, const SStruct &sstruct){
  const VI &mapping = sstruct.GetMapping();
  vector<double> posteriors (sstruct.GetLength()+1, 1.0);
  if (filename == "") return posteriors;

  ifstream infile (filename.c_str());
  if (infile.fail()) Error ("Could not read posteriors file.");

  string s;
  int length = 0;
  while (getline (infile, s)){
    istringstream iss (s);

    char let;
    int from;
    if (iss >> from >> let){
      length++;
      if (from != length) Error ("Bad nucleotide numbering in posteriors file.");
      if (!isalpha (let)) Error ("Bad letter in posteriors file.");
      while (iss >> s){
        string::size_type idx = s.find (':');
        if (idx == string::npos) Error ("Badly formatted line in posteriors file.");
        int to = atoi (s.substr (0, idx).c_str());
        if (to < 0) Error ("Negative mapping indices not allowed in posteriors file.");
        double value = atof (s.substr (idx + 1).c_str());
        if (value < -0.10 || value > 1.10) Error ("Invalid value in posteriors file.");
        value = max (0.0, min (1.0, value));
	
	if (from < 1 || from > sstruct.GetLength()) Error ("Index in posteriors file does not match BPSEQ length.");
	if (to < 1 || to > sstruct.GetLength()) Error ("Index in posteriors file does not match BPSEQ length.");
	if (mapping[from] == to) posteriors[from] = value;
	if (mapping[to] == from) posteriors[to] = value;
	if (mapping[from] == 0) posteriors[from] -= value;
	if (mapping[to] == 0) posteriors[to] -= value;
      }
    } else {
      Error ("Badly formatted line in posteriors file.");
    }
  }
  
  infile.close();
  
  return posteriors;
}

/////////////////////////////////////////////////////////////////
// DrawRNA()
//
// Draw RNA secondary structure.
/////////////////////////////////////////////////////////////////

void DrawRNA (const string &filename,
	      const SStruct &sstruct,
	      vector<Point> coords,
              const vector<double> &posteriors,
	      string title){

  const string &sequence = sstruct.GetSequence();
  const VI &mapping = sstruct.GetMapping();

  // transform and recenter coordinates
  
  for (int i = 1; i < (int) coords.size(); i++)
    coords[i].y *= -1;

  double min_x = numeric_limits<double>::infinity();
  double min_y = numeric_limits<double>::infinity();
  double max_x = -numeric_limits<double>::infinity();
  double max_y = -numeric_limits<double>::infinity();
  
  for (int i = 1; i < (int) coords.size(); i++){
    min_x = min (min_x, coords[i].x);
    min_y = min (min_y, coords[i].y);
    max_x = max (max_x, coords[i].x);
    max_y = max (max_y, coords[i].y);
  }
   
  int width = 0;
  int height = 0;
  for (int i = 1; i < (int) coords.size(); i++){
    coords[i].x = floor((coords[i].x - min_x) * SCALE);
    coords[i].y = floor((coords[i].y - min_y) * SCALE);
    width = max (width, (int) coords[i].x+1);
    height = max (height, (int) coords[i].y+1);
  }

  // create image

  int image_width = width + 2*BORDER;
  int image_height = height + 2*BORDER + HEADER;
  gdImagePtr image = gdImageCreateTrueColor (image_width, image_height);

  // allocate palette

  int white = gdImageColorAllocate (image, 255, 255, 255);
  int black = gdImageColorAllocate (image, 0, 0, 0);

  vector<int> confidence (101);
  for (int i = 0; i <= 100; i++){
    int r = (LOW_CONFIDENCE_COLOR[0] * (100 - i) + HIGH_CONFIDENCE_COLOR[0] * i) / 100;
    int g = (LOW_CONFIDENCE_COLOR[1] * (100 - i) + HIGH_CONFIDENCE_COLOR[1] * i) / 100;
    int b = (LOW_CONFIDENCE_COLOR[2] * (100 - i) + HIGH_CONFIDENCE_COLOR[2] * i) / 100;
    confidence[i] = gdImageColorAllocate (image, r, g, b);
  }

  // set white background

  gdImageFilledRectangle (image, 0, 0, image_width-1, image_height-1, white);

  // obtain font

  gdFontPtr text_font = gdFontGetMediumBold();
  gdFontPtr number_font = gdFontGetSmall();

  // draw RNA
  
  for (int i = 1; i < (int) coords.size(); i++){

    // choose color
    
    double p = posteriors[i];
    int index = max(0, min (100, (int)(p * 100)));
    
    // draw letter
  
    int x = BORDER + (int) coords[i].x;
    int y = HEADER + BORDER + (int) coords[i].y;

    gdImageSetAntiAliased (image, confidence[index]);
    gdImageChar (image, text_font, x - text_font->w/2, y - text_font->h/2, sequence[i], confidence[index]);

    // draw number if needed
    
    if (i % 10 == 0){
      string number = SPrintF ("%d", i);
      int lx = BORDER + (int) coords[i-1].x - x;
      int ly = HEADER + BORDER + (int) coords[i-1].y - y;
      int rx = (i < (int) coords.size() - 1) ? BORDER + (int) coords[i+1].x - x : -lx;
      int ry = (i < (int) coords.size() - 1) ? HEADER + BORDER + (int) coords[i+1].y - y : -ly;
      double al = atan2((double)ly,(double)lx);
      double ar = atan2((double)ry,(double)rx);
      if (al > ar) al += 2 * PI;
      double an = (al + ar) / 2;
      int nx = (int)(1.5 * SCALE * cos (an));
      int ny = (int)(1.5 * SCALE * sin (an));
      
      gdImageString (image, number_font, 
                     x + nx - (int) number.length() * number_font->w/2, 
                     y + ny - number_font->h/2, 
                     (unsigned char *) number.c_str(), black);
    }

    // draw bonds
    
    if (mapping[i] > i){
      int ox = BORDER + (int) coords[mapping[i]].x;
      int oy = HEADER + BORDER + (int) coords[mapping[i]].y;
      
      char left = toupper(sequence[i]);
      char right = toupper(sequence[mapping[i]]);

      // draw line for a Watson-Crick pair
      
      if (left == 'A' && (right == 'T' || right == 'U') ||
	  left == 'C' && right == 'G' ||
	  left == 'G' && right == 'C' ||
	  (left == 'T' || left == 'U') && right == 'A'){
	
        gdImageLine (image, 
                     (int)(0.75*x + 0.25*ox), (int)(0.75*y + 0.25*oy),
                     (int)(0.25*x + 0.75*ox), (int)(0.25*y + 0.75*oy),
                     gdAntiAliased);
        
      } 

      // draw line for GU pair
      
      else if (left == 'G' && (right == 'T' || right == 'U') ||
	       (left == 'T' || left == 'U') && right == 'G'){
        
        gdImageFilledArc (image,
                          (int)(0.5*x + 0.5*ox), (int)(0.5*y + 0.5*oy),
                          5, 5, 0, 360, gdAntiAliased, gdArc);
        
      }
    }
  }
  
  if (title != ""){

    // write title
    
    int max_chars = (int)(0.75 * image_width / number_font->w);
    if ((int) title.length() > max_chars){
      title = title.substr (0, max(0, max_chars - 3)) + "...";
    }
    
    gdImageString (image, number_font, image_width / 2 - (int) title.length() * number_font->w/2,
                   HEADER/2, (unsigned char *) title.c_str(), black);
  }
    
  // write to output file

  FILE *pngout = fopen (filename.c_str(), "wb");
  gdImagePng (image, pngout);
  fclose (pngout);
  gdImageDestroy (image);
  
}

/////////////////////////////////////////////////////////////////
// main()
//
// Main program.
/////////////////////////////////////////////////////////////////

int main (int argc, char **argv){

  if (argc < 4){
    cerr << "Usage: " << argv[0] << " [--posteriors POSTERIORSFILE] [--title TITLE] BPSEQFILE COORDSFILE OUTPUTPNGFILE" << endl;
    exit (1);
  }

  // parse arguments

  string posteriors_filename;
  string title;
  
  vector<string> default_args;
  for (int i = 1; i < argc; i++){
    if (argv[i][0] == '-'){
      if (string(argv[i]) == "--posteriors"){
        if (i == argc-1) Error ("Argument required after --posteriors.");
        posteriors_filename = argv[++i];
      } else if (string(argv[i]) == "--title"){
        if (i == argc-1) Error ("Argument required after --title.");
        title = argv[++i];
        HEADER = 50;
      } else {
        cerr << "Unknown argument: " << argv[i] << endl;
        exit (1);
      }
    } else {
      default_args.push_back (argv[i]);
    }
  }
  
  if (default_args.size() != 3) Error ("Incorrect number of arguments.");

  SStruct sstruct (default_args[0]);
  vector<Point> coords = ReadCoords (default_args[1]);
  vector<double> posteriors = ReadPosteriors (posteriors_filename, sstruct);
  
  DrawRNA (default_args[2], sstruct, coords, posteriors, title);
}
