/////////////////////////////////////////////////////////////////
// PlotRNA.h
/////////////////////////////////////////////////////////////////

#include "gd.h"
#include "gdfontmb.h"
#include "gdfonts.h"
#include <iostream>
#include <vector>
#include <map>
#include <string>
#include <sstream>
#include <fstream>
#include <cstring>
#include <cstdio>
#include <cmath>

using namespace std;

/////////////////////////////////////////////////////////////////
// Constants
/////////////////////////////////////////////////////////////////

const int HIGH_CONFIDENCE_COLOR[] = { 0, 0, 0 };
const int LOW_CONFIDENCE_COLOR[] = { 216, 216, 255 };

const int BORDER = 50;
const int EDGE = 0;
const int FONTSIZE = 14;
const double PI = 3.141592653589793238462643383279502884197169399;

/////////////////////////////////////////////////////////////////
// Globals
/////////////////////////////////////////////////////////////////

int HEADER = 0;

bool toggle_posteriors = false;
string title;

/////////////////////////////////////////////////////////////////
// Error()
//
// Report error and terminate program.
/////////////////////////////////////////////////////////////////

void Error (const string &s){
  cerr << "ERROR: " << s << endl;
  exit (1);
}

/////////////////////////////////////////////////////////////////
// ReadBPSEQ()
//
// Read BPSEQ file.
/////////////////////////////////////////////////////////////////

vector<pair<char,int> > ReadBPSEQ (const string &filename){
  
  vector<pair<char,int> > res (1, make_pair('@',0));
  ifstream infile (filename.c_str());
  if (infile.fail()) Error ("Could not read BPSEQ file.");

  string s;
  while (getline (infile, s)){
    istringstream iss (s);

    char let;
    int from, to;
    if (iss >> from >> let >> to){
      if (from != (int) res.size()) Error ("Bad nucleotide numbering in BPSEQ file.");
      if (!isalpha (let)) Error ("Bad letter in BPSEQ file.");
      if (to < 0) Error ("Negative mapping indices not allowed in BPSEQ file.");
      res.push_back (make_pair (let, to));
    } else {
      Error ("Badly formatted line in BPSEQ file.");
    }
  }
  
  infile.close();

  return res;
}

/////////////////////////////////////////////////////////////////
// ReadPosteriors()
//
// Read posteriors.
/////////////////////////////////////////////////////////////////

vector<vector<double> > ReadPosteriors (const string &filename){
  
  map<int, map<int, double> > posteriors;
  
  ifstream infile (filename.c_str());
  if (infile.fail()) Error ("Could not read posteriors file.");

  string s;
  int length = 0;
  while (getline (infile, s)){
    istringstream iss (s);

    char let;
    int from;
    if (iss >> from >> let){
      length++;
      if (from != length) Error ("Bad nucleotide numbering in posteriors file.");
      if (!isalpha (let)) Error ("Bad letter in posteriors file.");
      while (iss >> s){
        int idx = s.find (':');
        if (idx == -1) Error ("Badly formatted line in posteriors file.");
        int to = atoi (s.substr (0, idx).c_str());
        if (to < 0) Error ("Negative mapping indices not allowed in posteriors file.");
        double value = atof (s.substr (idx + 1).c_str());
        if (value < -0.10 || value > 1.10) Error ("Invalid value in posteriors file.");
        value = max (0.0, min (1.0, value));
        posteriors[from][to] = posteriors[to][from] = value;
      }
    } else {
      Error ("Badly formatted line in posteriors file.");
    }
  }
  
  infile.close();
  
  vector<vector<double> > res (length+1, vector<double>(length+1));
  
  for (map<int, map<int, double> >::iterator iter = posteriors.begin(); iter != posteriors.end(); ++iter){
    for (map<int, double>::iterator iter2 = iter->second.begin(); iter2 != iter->second.end(); ++iter2){
      res[iter->first][iter2->first] = iter2->second;
    }
  }
    
  return res;
}

/////////////////////////////////////////////////////////////////
// ReadCoords()
//
// Read coordinates file.
/////////////////////////////////////////////////////////////////

vector<pair<double,double> > ReadCoords (const string &filename){
  
  vector<pair<double,double> > res (1, make_pair(0,0));
  ifstream infile (filename.c_str());
  if (infile.fail()) Error ("Could not read coordinates file.");

  double first, second;
  while (infile >> first >> second){
    res.push_back (make_pair(first, second));
  }
  
  infile.close();
  return res;  
}

/////////////////////////////////////////////////////////////////
// DrawRNA()
//
// Draw RNA secondary structure.
/////////////////////////////////////////////////////////////////

void DrawRNA (const string &filename,
              const double scale,
              const vector<pair<char,int> > &mapping,
              vector<pair<double,double> > coords,
              const vector<vector<double> > &posteriors){

  // transform and recenter coordinates

  for (int i = 1; i < (int) coords.size(); i++){
    //swap (coords[i].first, coords[i].second);
    coords[i].first *= -1;
    coords[i].second *= -1;
  }

  double min_x = numeric_limits<double>::infinity();
  double min_y = numeric_limits<double>::infinity();
  double max_x = -numeric_limits<double>::infinity();
  double max_y = -numeric_limits<double>::infinity();

  for (int i = 1; i < (int) coords.size(); i++){
    min_x = min (min_x, coords[i].first);
    min_y = min (min_y, coords[i].second);
    max_x = max (max_x, coords[i].first);
    max_y = max (max_y, coords[i].second);
  }
  
  for (int i = 1; i < (int) coords.size(); i++){
    coords[i].first -= min_x;
    coords[i].second -= min_y;
  }

  double width = max_x - min_x;
  double height = max_y - min_y;

  for (int i = 1; i < (int) coords.size(); i++){
    coords[i].first = width - coords[i].first;
  }

  // create image

  int image_width = (int)(width * scale + 2*BORDER);
  int image_height = (int)(height * scale + 2*BORDER + HEADER);
  
  gdImagePtr image = gdImageCreateTrueColor (image_width, image_height);

  // allocate palette

  int white = gdImageColorAllocate (image, 255, 255, 255);
  int black = gdImageColorAllocate (image, 0, 0, 0);

  vector<int> confidence (101);
  for (int i = 0; i <= 100; i++){
    int r = toggle_posteriors ? (LOW_CONFIDENCE_COLOR[0] * (100 - i) + HIGH_CONFIDENCE_COLOR[0] * i) / 100 : 0;
    int g = toggle_posteriors ? (LOW_CONFIDENCE_COLOR[1] * (100 - i) + HIGH_CONFIDENCE_COLOR[1] * i) / 100 : 0;
    int b = toggle_posteriors ? (LOW_CONFIDENCE_COLOR[2] * (100 - i) + HIGH_CONFIDENCE_COLOR[2] * i) / 100 : 0;
    confidence[i] = gdImageColorAllocate (image, r, g, b);
  }

  // set white background

  gdImageFilledRectangle (image, EDGE, EDGE, image_width - 1 - EDGE, image_height - 1 - EDGE, white);  

  // obtain font

  gdFontPtr font = gdFontGetMediumBold();
  gdFontPtr number_font = gdFontGetSmall();

  // draw RNA
  
  for (int i = 1; i < (int) coords.size(); i++){

    // choose color
    
    double p = 0.0;
    if (mapping[i].second != 0){
      p = posteriors[i][mapping[i].second];
    } else {
      p = 1.0;
      for (int j = 1; j < (int) mapping.size(); j++)
        p -= posteriors[i][j];
    }
    int index = max(0, min (100, (int)(p * 100)));
    
    // draw letter
  
    int x = BORDER + (int)(coords[i].first * scale);
    int y = HEADER + BORDER + (int)(coords[i].second * scale);

    gdImageSetAntiAliased (image, confidence[index]);
    gdImageChar (image, font, x - font->w/2, y - font->h/2, mapping[i].first, confidence[index]);

    // draw number if needed
    
    if (i % 10 == 0){
      char number[10];
      sprintf (number, "%d", i);
      int lx = BORDER + (int)(coords[i-1].first * scale) - x;
      int ly = HEADER + BORDER + (int)(coords[i-1].second * scale) - y;
      int rx = (i < (int) coords.size() - 1) ? BORDER + (int)(coords[i+1].first * scale) - x : -lx;
      int ry = (i < (int) coords.size() - 1) ? HEADER + BORDER + (int)(coords[i+1].second * scale) - y : -ly;
      double al = atan2((double)ly,(double)lx);
      double ar = atan2((double)ry,(double)rx);
      if (al > ar) al += 2 * PI;
      double an = (al + ar) / 2;
      int nx = (int)(1.5 * scale * cos (an));
      int ny = (int)(1.5 * scale * sin (an));
      
      gdImageString (image, number_font, 
                     x + nx - (int) strlen(number) * number_font->w/2, 
                     y + ny - number_font->h/2, 
                     (unsigned char *) number, black);
    }

    // draw bonds
    
    if (mapping[i].second > i){
      int ox = BORDER + (int)(coords[mapping[i].second].first * scale);
      int oy = HEADER + BORDER + (int)(coords[mapping[i].second].second * scale);
      
      if (toupper(mapping[i].first) == 'A' && (toupper(mapping[mapping[i].second].first) == 'T' || toupper(mapping[mapping[i].second].first) == 'U') ||
          toupper(mapping[i].first) == 'C' && toupper(mapping[mapping[i].second].first) == 'G' ||
          toupper(mapping[i].first) == 'G' && toupper(mapping[mapping[i].second].first) == 'C' ||
          (toupper(mapping[i].first) == 'T' || toupper(mapping[i].first) == 'U') && toupper(mapping[mapping[i].second].first) == 'A'){

        gdImageLine (image, 
                     (int)(0.75*x + 0.25*ox), (int)(0.75*y + 0.25*oy),
                     (int)(0.25*x + 0.75*ox), (int)(0.25*y + 0.75*oy),
                     gdAntiAliased);
        
      } else if (toupper(mapping[i].first) == 'G' && (toupper(mapping[mapping[i].second].first) == 'T' || toupper(mapping[mapping[i].second].first) == 'U') ||
                 (toupper(mapping[i].first) == 'T' || toupper(mapping[i].first) == 'U') && toupper(mapping[mapping[i].second].first) == 'G'){
        
        gdImageFilledArc (image,
                          (int)(0.5*x + 0.5*ox), (int)(0.5*y + 0.5*oy),
                          5, 5, 0, 360, gdAntiAliased, gdArc);
        
      }
    }
  }

  if (title != ""){

    // write title
    
    int max_chars = (int)(0.75 * image_width / number_font->w);
    if ((int) title.length() > max_chars){
      title = title.substr (0, max(0, max_chars - 3));
      if (title.length() > 0) title += "...";
    }
    
    gdImageString (image, number_font, image_width / 2 - (int) title.length() * number_font->w/2,
                   HEADER/2, (unsigned char *) title.c_str(), black);
  }
    
  // write to output file

  FILE *pngout = fopen (filename.c_str(), "wb");
  gdImagePng (image, pngout);
  fclose (pngout);
  gdImageDestroy (image);
  
}

/////////////////////////////////////////////////////////////////
// main()
//
// Main program.
/////////////////////////////////////////////////////////////////

int main (int argc, char **argv){

  if (argc < 5){
    cerr << "Usage: plotRNA [--posteriors POSTERIORSFILE] [--title TITLE] SCALE BPSEQFILE COORDSFILE OUTPUTPNGFILE" << endl;
    exit (1);
  }

  // parse arguments

  string posterior_filename;
  
  vector<string> default_args;
  for (int i = 1; i < argc; i++){
    if (argv[i][0] == '-'){
      if (string(argv[i]) == "--posteriors"){
        if (i == argc-1) Error ("Argument required after --posteriors.");
        toggle_posteriors = true;
        posterior_filename = argv[++i];
      } else if (string(argv[i]) == "--title"){
        if (i == argc-1) Error ("Argument required after --title.");
        title = argv[++i];
        HEADER = 50;
      } else {
        cerr << "Unknown argument: " << argv[i] << endl;
        exit (1);
      }
    } else {
      default_args.push_back (argv[i]);
    }
  }
  
  if (default_args.size() != 4) Error ("Not enough filenames entered.");


  double scale = atof (default_args[0].c_str());
  vector<pair<char,int> > mapping = ReadBPSEQ(default_args[1].c_str());
  vector<pair<double,double> > coords = ReadCoords(default_args[2].c_str());

  vector<vector<double> > posteriors (mapping.size(), vector<double>(mapping.size(), 1));
  if (posterior_filename != "") posteriors = ReadPosteriors (posterior_filename);
  
  DrawRNA (default_args[3].c_str(), scale, mapping, coords, posteriors);
}
