/////////////////////////////////////////////////////////////////
// SStruct.cc
/////////////////////////////////////////////////////////////////

#include "SStruct.h"

extern bool toggle_complementary_only;

/////////////////////////////////////////////////////////////////
// SStruct::IsComplementary()
//
// Check if a base-pairing is one of AU, CG, or GU.
/////////////////////////////////////////////////////////////////

bool SStruct::IsComplementary (char c, char d){
  c = toupper(c);
  d = toupper(d);

  return 
    (c == 'A' && d == 'U') ||
    (c == 'U' && d == 'A') ||
    (c == 'C' && d == 'G') ||
    (c == 'G' && d == 'C') ||
    (c == 'G' && d == 'U') ||
    (c == 'U' && d == 'G');
}

/////////////////////////////////////////////////////////////////
// SStruct::AssertValidSequence()
//
// Check that a string represents a valid sequence.
/////////////////////////////////////////////////////////////////

void SStruct::AssertValidSequence (const string &sequence) const {
  for (int i = 1; i < (int) sequence.length(); i++){
    char ch = toupper (sequence[i]);
    if (ch != 'A' && ch != 'C' && ch != 'G' && ch != 'U' && ch != 'N')
      Error("Nucleotide %d has invalid type '%c'.", i, ch);
  }
}

/////////////////////////////////////////////////////////////////
// SStruct::AssertValidMapping()
//
// Check that a vector<int> represents a valid secondary
// structure mapping.  Pseudoknots are allowed.
/////////////////////////////////////////////////////////////////

void SStruct::AssertValidMapping (const vector<int> &mapping) const {
  if (mapping.size() == 0) return;
  for (int i = 1; i < (int) mapping.size(); i++){
    if (mapping[i] != UNPAIRED && mapping[i] != UNKNOWN){
      if (mapping[i] < 1 || mapping[i] >= (int) mapping.size())
        Error("Position %d of sequence maps to invalid position.", i);
      if (mapping[mapping[i]] != i)
        Error("Positions %d and %d of sequence do not map to each other.", i, mapping[i]);
      if (mapping[i] == i)
        Error("Position %d of sequence maps to itself.", i);
    }
  }
}

/////////////////////////////////////////////////////////////////
// SStruct::SStruct()
//
// Create object from file.  Attempt to detect the format of the
// file automatically.
/////////////////////////////////////////////////////////////////

SStruct::SStruct (const string &filename) : 
  sequence("@"), mapping(1, UNKNOWN), name("unnamed"){

  ifstream data (filename.c_str());
  if (data.fail()) Error (("Unable to open input file: " + filename).c_str());

  // look for first blank line
  
  string s;
  while (getline (data, s)){
    if (s.length() > 0) break;
  }
  
  // analyze to determine file format

  FileFormat format;
  if (s[0] == '>'){
    format = FileFormat_FASTA;
  } else {
    istringstream iss (s);
    int x, y;
    string t;
    if (iss >> x >> t >> y){
      format = FileFormat_BPSEQ;
    } else {
      format = FileFormat_RAW;
    }
  }
  
  data.close();

  // now load file

  switch (format){
  case FileFormat_FASTA: LoadFASTA (filename); break;
  case FileFormat_RAW: LoadRAW (filename); break;
  case FileFormat_BPSEQ: LoadBPSEQ (filename); break;
  default: Assert (false, "Should not get here.");
  }
  
  // convert T's to U's
  
  for (int i = 0; i < (int) sequence.length(); i++){
    if (sequence[i] == 't') sequence[i] = 'u';
    if (sequence[i] == 'T') sequence[i] = 'U';
  }

  AssertValidSequence (sequence);
  AssertValidMapping (mapping);
}
  
/////////////////////////////////////////////////////////////////
// SStruct::LoadFASTA()
//
// Create object from a FASTA file.
/////////////////////////////////////////////////////////////////

void SStruct::LoadFASTA (const string &filename){

  string s;
  
  ifstream data (filename.c_str());
  if (data.fail()) Error (("Unable to open input file: " + filename).c_str());

  // retrieve sequence header

  while (getline (data, s)){
    if (s.length() == 0) continue;
    Assert (s[0] == '>', "Auto-detection returned incorrect result.");
    name = s.substr(1);
    if (s.length() > 0) break;
  }

  // now retrieve sequence data
  
  while (getline (data, s)){
    if (s.length() == 0) continue;
    for (int i = 0; i < (int) s.length(); i++){
      sequence.push_back (s[i]);
      mapping.push_back (UNKNOWN);
    }
  }  
}

/////////////////////////////////////////////////////////////////
// SStruct::LoadRAW()
//
// Create object from raw (unformatted) file.
/////////////////////////////////////////////////////////////////

void SStruct::LoadRAW (const string &filename){

  string s;
  
  ifstream data (filename.c_str());
  if (data.fail()) Error (("Unable to open input file: " + filename).c_str());

  // now retrieve sequence data
  
  while (getline (data, s)){
    if (s.length() == 0) continue;
    for (int i = 0; i < (int) s.length(); i++){
      sequence.push_back (s[i]);
      mapping.push_back (UNKNOWN);
    }
  }  
}

/////////////////////////////////////////////////////////////////
// SStruct::LoadBPSEQ()
//
// Create object from BPSEQ file.
/////////////////////////////////////////////////////////////////

void SStruct::LoadBPSEQ (const string &filename){

  int index, maps_to;
  char ch;
  
  ifstream data (filename.c_str());
  if (data.fail()) Error (("Unable to open input file: " + filename).c_str());
  
  while (data >> index >> ch >> maps_to){
    if (index < 1) Error ("Only positive sequence indices allowed.");
    if (index != (int) mapping.size()) Error ("Sequence indices must appear in sorted order.");
    sequence.push_back (ch);
    mapping.push_back (maps_to);
  }
}

/////////////////////////////////////////////////////////////////
// SStruct::SStruct()
//
// Create object from given sequence, secondary structure
// mapping, and name.
/////////////////////////////////////////////////////////////////

SStruct::SStruct (const string &sequence, 
		  const vector<int> &mapping, 
		  const string &name) :
  sequence(sequence), mapping(mapping), name (name){

  Assert (sequence.length() > 0 && sequence[0] == '@', "Invalid sequence data format.");
  Assert (mapping.size() > 0 && mapping[0] == UNKNOWN, "Invalid mapping format.");
  Assert (sequence.length() == mapping.size(), "Sequence and structure lengths do not match.");
  AssertValidSequence (sequence);
  AssertValidMapping (mapping);
}

/////////////////////////////////////////////////////////////////
// SStruct::SStruct()
//
// Create object from existing SStruct object.
/////////////////////////////////////////////////////////////////

SStruct::SStruct (const SStruct &rhs) : 
  sequence(rhs.sequence), mapping(rhs.mapping), name (rhs.name){

  Assert (sequence.length() > 0 && sequence[0] == '@', "Invalid sequence data format.");
  Assert (mapping.size() > 0 && mapping[0] == UNKNOWN, "Invalid mapping format.");
  Assert (sequence.length() == mapping.size(), "Sequence and structure lengths do not match.");
  AssertValidSequence (sequence);
  AssertValidMapping (mapping);
}

/////////////////////////////////////////////////////////////////
// SStruct::operator=
//
// Assignment operator.
/////////////////////////////////////////////////////////////////

const SStruct& SStruct::operator= (const SStruct &rhs){
  if (this != &rhs){
    sequence = rhs.sequence;
    mapping = rhs.mapping;
    name = rhs.name;
  }
  
  Assert (sequence.length() > 0 && sequence[0] == '@', "Invalid sequence data format.");
  Assert (mapping.size() > 0 && mapping[0] == UNKNOWN, "Invalid mapping format.");
  Assert (sequence.length() == mapping.size(), "Sequence and structure lengths do not match.");
  AssertValidSequence (sequence);
  AssertValidMapping (mapping);
  return *this;
}

/////////////////////////////////////////////////////////////////
// Getters
/////////////////////////////////////////////////////////////////

const string &SStruct::GetSequence() const { return sequence; }
const vector<int> &SStruct::GetMapping () const { return mapping; }
const string &SStruct::GetName () const { return name; }
int SStruct::GetLength () const { return mapping.size()-1; }

/////////////////////////////////////////////////////////////////
// SStruct::GetComplementaryMapping()
//
// Get mapping containing only complementary pairs.
/////////////////////////////////////////////////////////////////

vector<int> SStruct::GetComplementaryMapping () const {
  vector<int> new_mapping = mapping;

  for (int i = 1; i < (int) new_mapping.size(); i++){
    if (new_mapping[i] != UNPAIRED && new_mapping[i] != UNKNOWN && 
	!IsComplementary(sequence[i], sequence[new_mapping[i]])){
      new_mapping[i] = UNPAIRED;
    }
  }

  return new_mapping;
}

/////////////////////////////////////////////////////////////////
// SStruct::WriteBPSEQ()
//
// Write sequence in BPSEQ format.
/////////////////////////////////////////////////////////////////

void SStruct::WriteBPSEQ (ostream &outfile) const {
  vector<int> new_mapping = toggle_complementary_only ? GetComplementaryMapping() : mapping;

  for (int i = 1; i < (int) new_mapping.size(); i++)
    outfile << i << " " << sequence[i] << " " << new_mapping[i] << endl;
}

/////////////////////////////////////////////////////////////////
// SStruct::WriteParens()
//
// Write sequence in parenthesized format.  This routine assumes
// that the structure does not contain pseudoknots.
/////////////////////////////////////////////////////////////////

void SStruct::WriteParens (ostream &outfile) const {
  vector<int> new_mapping = toggle_complementary_only ? GetComplementaryMapping() : mapping;

  outfile << ">" << name << endl;
  for (int i = 1; i < (int) sequence.length(); i++)
    outfile << sequence[i];
  outfile << endl;
  for (int i = 1; i < (int) new_mapping.size(); i++){
    if (new_mapping[i] == UNKNOWN){
      outfile << "?";
    } else if (new_mapping[i] == UNPAIRED){
      outfile << ".";
    } else if (new_mapping[i] > i){
      outfile << "(";
    } else {
      outfile << ")";
    }
  }
  outfile << endl;
}
