/////////////////////////////////////////////////////////////////
// Table.cc
/////////////////////////////////////////////////////////////////

#include "Table.h"

extern bool toggle_complementary_only;
extern double posterior_cutoff;

/////////////////////////////////////////////////////////////////
// Table::Table()
//
// Constructor.
/////////////////////////////////////////////////////////////////

Table::Table (int layers, int size) : layers(layers), size(size) {
  Assert (layers > 0, "Number of layers in matrix must be positive.");
  Assert (size > 0, "Number of rows in matrix must be positive.");

  data = VVVF (layers, VVF (size, VF (size)));
}

/////////////////////////////////////////////////////////////////
// Table::Table()
//
// Copy constructor.
/////////////////////////////////////////////////////////////////

Table::Table (const Table &rhs) : layers (rhs.layers), size(rhs.size){
  Assert (layers > 0, "Number of layers in matrix must be positive.");
  Assert (size > 0, "Number of rows in matrix must be positive.");
}

/////////////////////////////////////////////////////////////////
// Table::Fill()
//
// Fill in matrix values.
/////////////////////////////////////////////////////////////////

void Table::Fill (const Score &value){
  for (int i = 0; i < layers; i++){
    for (int j = 0; j < size; j++){
      for (int k = j; k < size; k++){
	data[i][j][k] = value;
      }
    }
  }
}

/////////////////////////////////////////////////////////////////
// Table::PrintVal()
//
// Printing utility function.
/////////////////////////////////////////////////////////////////

void Table::PrintVal (ostream &outfile, const Score &value) const {
  if (value == LOG_ZERO)
    outfile << setw(9) << " -inf";
  else
    outfile << setw(9) << setprecision(6) << setiosflags(ios::fixed | ios::showpoint) 
	    << value << resetiosflags(ios::fixed | ios::showpoint);
}

/////////////////////////////////////////////////////////////////
// Table::PrintVal()
//
// Print a single matrix layer.
/////////////////////////////////////////////////////////////////

void Table::PrintLayer (ostream &outfile, int layer) const {

  for (int i = 0; i < min(12,size); i++){
    outfile << (i == 0 ? "[" : " ") << "[";
    for (int j = 0; j < min(12,size); j++){
      if (j > 0) outfile << ", ";
      if (i <= j)
	PrintVal (outfile, operator()(layer,i,j));
      else
	PrintVal (outfile, LOG_ZERO);
    }
    outfile << "]" << (i == size-1 ? "]" : "") << endl;
  }
  outfile << endl;
}

/////////////////////////////////////////////////////////////////
// Table::Print()
//
// Print all matrix layers
/////////////////////////////////////////////////////////////////

void Table::Print (ostream &outfile) const {
  for (int i = 0; i < layers; i++)
    PrintLayer (outfile, i);
}

/////////////////////////////////////////////////////////////////
// Table::PrintSparse()
//
// Print sparse matrix.
/////////////////////////////////////////////////////////////////

void Table::PrintSparse (ostream &outfile, const string &seq) const {
  Assert (size == seq.length(), "Table size does not match sequence length.");

  for (int i = 1; i < size; i++){
    outfile << i << " " << seq[i];
    for (int j = i+1; j < size; j++){
      const Score value = operator()(0,i,j);
      if (value < posterior_cutoff) continue;
      if (toggle_complementary_only && !SStruct::IsComplementary (seq[i], seq[j])) continue;
      outfile << " " << j << ":" << setprecision(3) << setiosflags(ios::fixed | ios::showpoint)
	      << value << resetiosflags(ios::fixed | ios::showpoint);
    }
    outfile << endl;
  }
}
